# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.types import PackageCategories
from rebasehelper.results_store import results_store


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class BaseBuildLogHook(Plugin):
    """Base class for a build log hook."""

    CATEGORIES: PackageCategories = []

    @classmethod
    def format(cls, data):
        """Formats build log hook output to a readable text form."""
        raise NotImplementedError()

    @classmethod
    def run(cls, spec_file, rebase_spec_file, results_dir, **kwargs):
        """Runs the build log hook.

        Args:
            spec_file (rebasehelper.specfile.SpecFile): Original SpecFile object.
            rebase_spec_file (rebasehelper.specfile.SpecFile): Rebased SpecFile object.
            kwargs (dict): Keyword arguments from instance of Application.

        Returns:
            tuple: The first element is a dict containing changes, the second is a bool whether
            the build process should be restarted.

        """
        raise NotImplementedError()

    @classmethod
    def merge_two_results(cls, old, new):
        raise NotImplementedError()


class BuildLogHookCollection(PluginCollection):
    def run(self, spec_file, rebase_spec_file, non_interactive, force_build_log_hooks, **kwargs):
        """Runs all non-blacklisted build log hooks.

        Args:
            spec_file (rebasehelper.specfile.SpecFile): Original SpecFile object.
            rebase_spec_file (rebasehelper.specfile.SpecFile): Rebased SpecFile object.
            non_interactive (bool): Whether rebase-helper is in non-interactive mode.
            force_build_log_hooks (bool): Whether to run the hooks even in
                non-interactive mode.
            kwargs (dict): Keyword arguments from instance of Application.

        Returns:
            bool: Whether build log hooks made some changes to the SPEC file.

        """
        changes_made = False
        if not non_interactive or force_build_log_hooks:
            blacklist = kwargs.get('build_log_hook_blacklist', [])
            for name, build_log_hook in self.plugins.items():
                if not build_log_hook or name in blacklist:
                    continue
                categories = build_log_hook.CATEGORIES
                if not categories or spec_file.category in categories:
                    logger.info('Running %s build log hook.', name)
                    result, rerun = build_log_hook.run(spec_file, rebase_spec_file, **kwargs)
                    result = build_log_hook.merge_two_results(results_store.get_build_log_hooks().get(name, {}), result)
                    results_store.set_build_log_hooks_result(name, result)
                    if rerun:
                        changes_made = True
        return changes_made



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import datetime
import fcntl
import io
import os
import re
import sys
import tempfile
import termios
import tty

import colors  # type: ignore

from rebasehelper.constants import SYSTEM_ENCODING


class ConsoleHelper:

    """Class for interacting with the command line."""

    use_colors: bool = False

    @classmethod
    def should_use_colors(cls, conf):
        """Determines whether ANSI colors should be used for CLI output.

        Args:
            conf (rebasehelper.config.Config): Configuration object with arguments from the command line.

        Returns:
            bool: Whether colors should be used.

        """
        if os.environ.get('PY_COLORS') == '1':
            return True
        if os.environ.get('PY_COLORS') == '0':
            return False
        if conf.color == 'auto':
            if (not os.isatty(sys.stdout.fileno()) or
                    os.environ.get('TERM') == 'dumb'):
                return False
        elif conf.color == 'never':
            return False
        return True

    @classmethod
    def cprint(cls, message, fg=None, bg=None, style=None):
        """Prints colored output if possible.

        Args:
            message (str): String to be printed out.
            fg (str): Foreground color.
            bg (str): Background color.
            style (str): Style to be applied to the printed message.
                Possible styles: bold, faint, italic, underline, blink, blink2, negative, concealed, crossed.
                Some styles may not be supported by every terminal, e.g. 'blink'.
                Multiple styles should be connected with a '+', e.g. 'bold+italic'.

        """
        if cls.use_colors:
            try:
                print(colors.color(message, fg=fg, bg=bg, style=style))
            except ValueError:
                print(message)
        else:
            print(message)

    @staticmethod
    def parse_rgb_device_specification(specification):
        """Parses RGB device specification.

        Args:
            specification(str): RGB device specification.

        Returns:
            tuple: If the specification follows correct format, the first element is RGB tuple and the second is
            bit width of the RGB. Otherwise, both elements are None.

        """
        match = re.match(r'^rgb:([A-Fa-f0-9]{1,4})/([A-Fa-f0-9]{1,4})/([A-Fa-f0-9]{1,4})$', str(specification))
        if match:
            rgb = match.groups()
            bit_width = max(len(str(x)) for x in rgb) * 4
            return tuple(int(x, 16) for x in rgb), bit_width
        return None, None

    @staticmethod
    def color_is_light(rgb, bit_width):
        """Determines whether a color is light or dark.

        Args:
            rgb(tuple): RGB tuple.
            bit_width: Number of bits defining the RGB.

        Returns:
            bool: Whether a color is light or dark.

        """
        brightness = 1 - (0.299 * rgb[0] + 0.587 * rgb[1] + 0.114 * rgb[2]) / (1 << bit_width - 1)
        return brightness < 0.5

    @staticmethod
    def detect_background():
        """Detects terminal background color and decides whether it is light or dark.

        Returns:
            str: Whether to use dark or light color scheme.

        """
        background_color = ConsoleHelper.exchange_control_sequence('\x1b]11;?\x07')

        rgb_tuple, bit_width = ConsoleHelper.parse_rgb_device_specification(background_color)

        if rgb_tuple and ConsoleHelper.color_is_light(rgb_tuple, bit_width):
            return 'light'
        else:
            return 'dark'

    @staticmethod
    def exchange_control_sequence(query, timeout=0.05):
        """Captures a response of a control sequence from STDIN.

        Args:
            query (str): Control sequence.
            timeout (int, float): Time given to the terminal to react.

        Returns:
            str: Response of the terminal.

        """
        prefix, suffix = query.split('?', 1)
        attrs_obtained = False
        try:
            attrs = termios.tcgetattr(sys.stdin)
            attrs_obtained = True
            flags = fcntl.fcntl(sys.stdin.fileno(), fcntl.F_GETFL)

            # disable STDIN line buffering
            tty.setcbreak(sys.stdin.fileno(), termios.TCSANOW)
            # set STDIN to non-blocking mode
            fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK)

            sys.stdout.write(query)
            sys.stdout.flush()

            # read the response
            buf = ''
            start = datetime.datetime.now()
            while (datetime.datetime.now() - start).total_seconds() < timeout:
                try:
                    buf += sys.stdin.read(1)
                except IOError:
                    continue
                if buf.endswith(suffix):
                    return buf.replace(prefix, '').replace(suffix, '')
            return None
        except termios.error:
            return None
        finally:
            # set terminal settings to the starting point
            if attrs_obtained:
                termios.tcsetattr(sys.stdin, termios.TCSADRAIN, attrs)
                fcntl.fcntl(sys.stdin.fileno(), fcntl.F_SETFL, flags)

    class Capturer:
        """ContextManager for capturing stdout/stderr"""

        def __init__(self, stdout=False, stderr=False):
            self.capture_stdout = stdout
            self.capture_stderr = stderr
            self.stdout = None
            self.stderr = None
            self._stdout_fileno = None
            self._stderr_fileno = None
            self._stdout_tmp = None
            self._stderr_tmp = None
            self._stdout_copy = None
            self._stderr_copy = None

        def __enter__(self):
            self._stdout_fileno = sys.__stdout__.fileno()  # pylint: disable=no-member
            self._stderr_fileno = sys.__stderr__.fileno()  # pylint: disable=no-member

            self._stdout_tmp = tempfile.TemporaryFile(mode='w+b') if self.capture_stdout else None
            self._stderr_tmp = tempfile.TemporaryFile(mode='w+b') if self.capture_stderr else None
            self._stdout_copy = os.fdopen(os.dup(self._stdout_fileno), 'wb') if self.capture_stdout else None
            self._stderr_copy = os.fdopen(os.dup(self._stderr_fileno), 'wb') if self.capture_stderr else None

            if self._stdout_tmp:
                sys.stdout.flush()
                os.dup2(self._stdout_tmp.fileno(), self._stdout_fileno)
            if self._stderr_tmp:
                sys.stderr.flush()
                os.dup2(self._stderr_tmp.fileno(), self._stderr_fileno)

            return self

        def __exit__(self, *args):
            if self._stdout_copy:
                sys.stdout.flush()
                os.dup2(self._stdout_copy.fileno(), self._stdout_fileno)
            if self._stderr_copy:
                sys.stderr.flush()
                os.dup2(self._stderr_copy.fileno(), self._stderr_fileno)

            if self._stdout_tmp:
                self._stdout_tmp.flush()
                self._stdout_tmp.seek(0, io.SEEK_SET)
                self.stdout = self._stdout_tmp.read().decode(SYSTEM_ENCODING)
            if self._stderr_tmp:
                self._stderr_tmp.flush()
                self._stderr_tmp.seek(0, io.SEEK_SET)
                self.stderr = self._stderr_tmp.read().decode(SYSTEM_ENCODING)

            if self._stdout_tmp:
                self._stdout_tmp.close()
            if self._stderr_tmp:
                self._stderr_tmp.close()
            if self._stdout_copy:
                self._stdout_copy.close()
            if self._stderr_copy:
                self._stderr_copy.close()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore
import rpm  # type: ignore

from rebasehelper.tags import Tag, Tags
from rebasehelper.spec_content import SpecContent


class TestTags:

    RAW_CONTENT = '''
Name: test
Version: 0.1
Source: test.tar.xz
SOURCE1: additional-files.zip
Patch00: patch0000.diff
Patch:   patch0001.diff
Patch:   patch0002.diff
Patch03: patch0003.diff
%{?!optimize:Patch999: extras.patch}
%{?optimize:Patch999: extras-optimized.patch}

%sourcelist
   # test comment
source2.zip

%patchlist
list.patch

%Package libs
Requires: %{name}-%{version}
Requires: extradep
Source: source3.tar.gz

%sourcelist
source4.zip

%if 0%{?with_utils:1}
%package utils
Requires: %{name}-%{version}
%endif
    '''

    PARSED_CONTENT = '''
Name: test
Version: 0.1
Source: test.tar.xz
SOURCE1: additional-files.zip
Patch00: patch0000.diff
Patch:   patch0001.diff
Patch:   patch0002.diff
Patch03: patch0003.diff

Patch999: extras-optimized.patch

%sourcelist
   # test comment
source2.zip

%patchlist
list.patch

%Package libs
Requires: test-0.1
Requires: extradep
Source: source3.tar.gz

%sourcelist
source4.zip

    '''

    MACROS = [
        ('name', 'test'),
        ('version', '0.1'),
        ('optimize', '1'),
    ]

    @pytest.fixture
    def tags(self):
        rpm.reloadConfig()
        for macro in self.MACROS:
            rpm.addMacro(*macro)
        return Tags(SpecContent(self.RAW_CONTENT), SpecContent(self.PARSED_CONTENT))

    def test_tags(self, tags):
        assert len(tags) == 16
        assert tags[0] == Tag(0, '%package', 1, 'Name', (6, 10), True)
        assert tags[1] == Tag(0, '%package', 2, 'Version', (9, 12), True)
        assert tags[2] == Tag(0, '%package', 3, 'Source0', (8, 19), True, 0)
        assert tags[3] == Tag(0, '%package', 4, 'Source1', (9, 29), True, 1)
        assert tags[4] == Tag(0, '%package', 5, 'Patch0', (9, 23), True, 0)
        assert tags[5] == Tag(0, '%package', 6, 'Patch1', (9, 23), True, 1)
        assert tags[6] == Tag(0, '%package', 7, 'Patch2', (9, 23), True, 2)
        assert tags[7] == Tag(0, '%package', 8, 'Patch3', (9, 23), True, 3)
        assert tags[8] == Tag(0, '%package', 10, 'Patch999', (22, 44), True, 999)
        assert tags[9] == Tag(1, '%sourcelist', 1, 'Source2', (0, 11), True, 2)
        assert tags[10] == Tag(2, '%patchlist', 0, 'Patch1000', (0, 10), True, 1000)
        assert tags[11] == Tag(3, '%package libs', 0, 'Requires', (10, 28), True)
        assert tags[12] == Tag(3, '%package libs', 1, 'Requires', (10, 18), True)
        assert tags[13] == Tag(3, '%package libs', 2, 'Source3', (8, 22), True, 3)
        assert tags[14] == Tag(4, '%sourcelist', 0, 'Source4', (0, 11), True, 4)
        assert tags[15] == Tag(5, '%package utils', 0, 'Requires', (10, 28), False)

    def test_filter(self, tags):
        assert len(list(tags.filter(section_name='%package'))) == 9
        assert len(list(tags.filter(section_name='%package', name='Patch*'))) == 5
        assert len(list(tags.filter(section_name='%package', valid=False))) == 0
        assert len(list(tags.filter(valid=False))) == 1
        assert len(list(tags.filter(name='Requires', valid=None))) == 3



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re
from typing import List, Dict, Type

from rebasehelper.types import Options
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.plugins.build_log_hooks import BuildLogHookCollection
from rebasehelper.plugins.build_tools.rpm import BuildToolCollection
from rebasehelper.plugins.build_tools.srpm import SRPMBuildToolCollection
from rebasehelper.plugins.checkers import CheckerCollection
from rebasehelper.plugins.output_tools import OutputToolCollection
from rebasehelper.plugins.spec_hooks import SpecHookCollection
from rebasehelper.plugins.versioneers import VersioneerCollection


class PluginManager:

    COLLECTIONS: List[Type[PluginCollection]] = [
        BuildLogHookCollection,
        BuildToolCollection,
        SRPMBuildToolCollection,
        CheckerCollection,
        OutputToolCollection,
        SpecHookCollection,
        VersioneerCollection,
    ]

    def __init__(self):
        def convert_class_name(class_name):
            # Converts class name of a collection to the corresponding entrypoint name,
            # e.g. BuildLogHookCollection => build_log_hooks
            class_name = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', class_name)
            return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', class_name).lower().replace('_collection', '') + 's'

        self.plugin_collections: Dict[str, PluginCollection] = {}
        for collection in self.COLLECTIONS:
            name = convert_class_name(collection.__name__)
            entrypoint = 'rebasehelper.' + name
            self.plugin_collections[name] = collection(entrypoint, self)

    def get_options(self) -> Options:
        """Gets options of all plugins.

        Returns:
            list: List of plugins' options.

        """
        options: Options = []
        for collection in self.plugin_collections.values():
            options.extend(collection.get_options())

        return options

    def __getattr__(self, name):
        return self.plugin_collections.get(name)


# Global instance of PluginManager, it is enough to load it once per application run.
plugin_manager: PluginManager = PluginManager()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import fnmatch
import os
import tempfile


class PathHelper:

    """Class for performing path related tasks."""

    @staticmethod
    def find_first_dir_with_file(top_path, pattern):
        """Recursively searches for a directory containing a file that matches the given pattern.

        Args:
            top_path (str): Directory where to start the search.
            pattern (str): Filename pattern.

        Returns:
            str: Full path to the directory containing the first occurence of the searched file.
            None if there is no file matching the pattern.

        """
        for root, dirs, files in os.walk(top_path):
            dirs.sort()
            for f in files:
                if fnmatch.fnmatch(f, pattern):
                    return os.path.abspath(root)
        return None

    @staticmethod
    def find_first_file(top_path, pattern, recursion_level=None):
        """Recursively searches for a file that matches the given pattern.

        Args:
            top_path (str): Directory where to start the search.
            pattern (str): Filename pattern.
            recursion_level (int): How deep in the directory tree the search can go.

        Returns:
            str: Path to the file matching the pattern or None if there is no file matching the pattern.

        """
        for loop, (root, dirs, files) in enumerate(os.walk(top_path)):
            dirs.sort()
            for f in files:
                if fnmatch.fnmatch(f, pattern):
                    return os.path.join(os.path.abspath(root), f)
            if recursion_level is not None:
                if loop == recursion_level:
                    break
        return None

    @staticmethod
    def find_all_files(top_path, pattern):
        """Recursively searches for all files matching the given pattern.

        Args:
            top_path (str): Directory where to start the search.
            pattern (str): Filename pattern.

        Returns:
            list: List containing absolute paths to all found files.

        """
        files_list = []
        for root, dirs, files in os.walk(top_path):
            dirs.sort()
            for f in files:
                if fnmatch.fnmatch(f, pattern):
                    files_list.append(os.path.join(os.path.abspath(root), f))
        return files_list

    @staticmethod
    def find_all_files_current_dir(top_path, pattern):
        """Searches for all files that match the given pattern inside a directory.

        Args:
            top_path (str): Directory where to start the search.
            pattern (str): Filename pattern.

        Returns:
            list: List containing absolute paths to all found files.

        """
        files_list = []
        for files in os.listdir(top_path):
            if fnmatch.fnmatch(files, pattern):
                files_list.append(os.path.join(os.path.abspath(top_path), files))
        return files_list

    @staticmethod
    def get_temp_dir():
        """Creates a new temporary directory.

        Return:
            str: Path to the created directory.

        """
        return tempfile.mkdtemp(prefix='rebase-helper-')

    @staticmethod
    def file_available(filename):
        """Checks if the given file exists.

        Args:
            filename (str): Path to the file.

        Returns:
            bool: Whether the file exists.

        """
        if os.path.exists(filename) and os.path.getsize(filename) != 0:
            return True
        else:
            return False



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import re
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.versioneers import BaseVersioneer
from rebasehelper.types import PackageCategories
from rebasehelper.specfile import PackageCategory
from rebasehelper.helpers.download_helper import DownloadHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class PyPI(BaseVersioneer):

    CATEGORIES: PackageCategories = [PackageCategory.python]

    BASE_URL: str = 'https://pypi.org'
    API_URL: str = '{}/pypi'.format(BASE_URL)

    @classmethod
    def _get_version(cls, package_name):
        r = DownloadHelper.request('{}/{}/json'.format(cls.API_URL, package_name))
        if r is None or not r.ok:
            # try to strip python prefix
            package_name = re.sub(r'^python[23]?-', '', package_name)
            r = DownloadHelper.request('{}/{}/json'.format(cls.API_URL, package_name))
            if r is None or not r.ok:
                return None

        data = r.json()
        try:
            return data['info']['version']
        except KeyError:
            return None

    @classmethod
    def run(cls, package_name):
        version = cls._get_version(package_name)
        if version:
            return version
        logger.error("Failed to determine latest upstream version!\n"
                     "Check that the package exists on %s.", cls.BASE_URL)
        return None



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import json

from rebasehelper.plugins.output_tools import BaseOutputTool
from rebasehelper.results_store import results_store


class JSON(BaseOutputTool):
    """ JSON output tool """

    EXTENSION: str = 'json'

    @classmethod
    def print_summary(cls, report_path, results):
        """Writes the report as JSON data.

        Args:
            report_path: Path to the report.
            results: Results store instance to get the data from.

        """
        with open(report_path, 'w') as outputfile:
            json.dump(results.get_all(), outputfile, indent=4, sort_keys=True)

    @classmethod
    def run(cls, logs, app):  # pylint: disable=unused-argument
        """Runs the output tool, writes the data to a JSON-formatted report."""
        cls.print_summary(cls.get_report_path(app), results_store)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import sys
from argparse import SUPPRESS
from typing import List

from rebasehelper.cli import CLI
from rebasehelper.constants import CONFIG_PATH, CONFIG_FILENAME


class SampleConfig:

    DESCRIPTION: List[str] = [
        '# sample configuration file for rebase-helper',
        '# copy this file to {} and edit it as necessary'.format(os.path.join(CONFIG_PATH, CONFIG_FILENAME)),
        '# all options specified here can be overridden on the command line',
    ]

    BLACKLIST: List[str] = [
        'help',
        'version',
        'config-file',
    ]

    @classmethod
    def generate(cls):
        result = cls.DESCRIPTION + ['']
        parser = CLI.build_parser()
        result.append('[general]')
        for action in parser._get_optional_actions():  # pylint: disable=protected-access
            if action.help is SUPPRESS:
                continue
            fmt = parser._get_formatter()  # pylint: disable=protected-access
            opts = action.option_strings
            if len(opts) > 1:
                opts.pop(0)
            name = opts[0].lstrip('-')
            if name in cls.BLACKLIST:
                continue
            value = getattr(action, 'actual_default', None)
            if isinstance(value, list):
                value = ','.join(value)
            args = fmt._format_args(action, action.dest.upper())  # pylint: disable=protected-access
            result.append('')
            result.append('# {}'.format(fmt._expand_help(action)))  # pylint: disable=protected-access
            if args:
                result.append('# synopsis: {} = {}'.format(name, args))
            result.append('{} = {}'.format(name, value if value is not None else ''))
        return '\n'.join(result)


def main():
    if len(sys.argv) != 2:
        return 1
    s = SampleConfig.generate()
    with open(sys.argv[1], 'w') as f:
        f.write(s)
    return 0


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import fnmatch
import logging
import os
import shutil
import urllib.parse
from typing import Any, Dict, List, Optional, cast

import git  # type: ignore
from pkg_resources import parse_version

from rebasehelper.archive import Archive
from rebasehelper.specfile import SpecFile, get_rebase_name
from rebasehelper.logger import CustomLogger, LoggerHelper
from rebasehelper import constants
from rebasehelper.config import Config
from rebasehelper.patcher import Patcher
from rebasehelper.plugins.plugin_manager import plugin_manager
from rebasehelper.plugins.checkers import CheckerCategory
from rebasehelper.exceptions import RebaseHelperError, CheckerNotFoundError
from rebasehelper.exceptions import SourcePackageBuildError, BinaryPackageBuildError
from rebasehelper.results_store import results_store
from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.helpers.macro_helper import MacroHelper
from rebasehelper.helpers.input_helper import InputHelper
from rebasehelper.helpers.git_helper import GitHelper
from rebasehelper.helpers.koji_helper import KojiHelper
from rebasehelper.helpers.lookaside_cache_helper import LookasideCacheHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Application:

    def __init__(self, cli_conf: Config, start_dir: str, execution_dir: str, results_dir: str,
                 create_logs: bool = True) -> None:
        """Initializes the application.

        Args:
            cli_conf: Application configuration.
            start_dir: Directory where rebase-helper was started.
            execution_dir: Working directory.
            results_dir: Location of rebase results.
            create_logs: Whether to create default logging file handlers.

        """
        results_store.clear()

        # Initialize instance attributes
        self.old_sources = ''
        self.new_sources = ''
        self.old_rest_sources: List[str] = []
        self.new_rest_sources: List[str] = []
        self.rebased_patches: Dict[str, List[str]] = {}
        self.rebased_repo: Optional[git.Repo] = None

        self.handlers = LoggerHelper.create_file_handlers(results_dir) if create_logs else []

        self.conf = cli_conf
        self.start_dir = start_dir
        self.execution_dir = execution_dir
        self.rebased_sources_dir = os.path.join(results_dir, 'rebased-sources')

        self.kwargs: Dict[str, Any] = {}
        self.kwargs.update(self.conf.config)
        # Temporary workspace for Builder, checks, ...
        workspace_location = os.path.abspath(cli_conf.workspace_dir) if cli_conf.workspace_dir else self.execution_dir
        self.kwargs['workspace_dir'] = self.workspace_dir = os.path.join(workspace_location, constants.WORKSPACE_DIR)

        # Directory where results should be put
        self.kwargs['results_dir'] = self.results_dir = results_dir
        # Directory contaning only those files, which are relevant for the new rebased version
        self.kwargs['rebased_sources_dir'] = self.rebased_sources_dir

        self.spec_file_path = self._find_spec_file()
        self._prepare_spec_objects()

        if self.conf.build_tasks is None:
            # check the workspace dir
            self._check_workspace_dir()

            # verify all sources for the new version are present
            missing_sources = [os.path.basename(s) for s in self.rebase_spec_file.sources
                               if not os.path.isfile(os.path.basename(s))]
            if missing_sources:
                raise RebaseHelperError('The following sources are missing: {}'.format(','.join(missing_sources)))

            if self.conf.update_sources:
                sources = [os.path.basename(s) for s in self.spec_file.sources]
                rebased_sources = [os.path.basename(s) for s in self.rebase_spec_file.sources]
                uploaded = LookasideCacheHelper.update_sources('fedpkg', self.rebased_sources_dir,
                                                               self.rebase_spec_file.header.name,
                                                               sources, rebased_sources,
                                                               upload=not self.conf.skip_upload)
                self._update_gitignore(uploaded, self.rebased_sources_dir)

            self._initialize_data()

    def __del__(self):
        LoggerHelper.remove_file_handlers(self.handlers)

    @staticmethod
    def setup(cli_conf):
        execution_dir = os.getcwd()
        results_dir = os.path.abspath(cli_conf.results_dir) if cli_conf.results_dir else execution_dir
        results_dir = os.path.join(results_dir, constants.RESULTS_DIR)

        # if not continuing, check the results dir
        Application._check_results_dir(results_dir)

        return execution_dir, results_dir

    def _prepare_spec_objects(self):
        """
        Prepare spec files and initialize objects

        :return:
        """
        self.spec_file = SpecFile(self.spec_file_path, self.execution_dir, self.kwargs['rpmmacros'])
        # Check whether test suite is enabled at build time
        if not self.spec_file.is_test_suite_enabled():
            results_store.set_info_text('WARNING', 'Test suite is not enabled at build time.')
        # create an object representing the rebased SPEC file
        self.rebase_spec_file = self.spec_file.copy(get_rebase_name(self.rebased_sources_dir, self.spec_file_path))

        if not self.conf.sources:
            self.conf.sources = plugin_manager.versioneers.run(self.conf.versioneer,
                                                               self.spec_file.header.name,
                                                               self.spec_file.category,
                                                               self.conf.versioneer_blacklist)
            if self.conf.sources:
                logger.info("Determined latest upstream version '%s'", self.conf.sources)
            else:
                raise RebaseHelperError('Could not determine latest upstream version '
                                        'and no SOURCES argument specified!')

        # Prepare rebased_sources_dir
        self.rebased_repo = self._prepare_rebased_repository()

        # check if argument passed as new source is a file or just a version
        if [True for ext in Archive.get_supported_archives() if self.conf.sources.endswith(ext)]:
            logger.verbose("argument passed as a new source is a file")
            version_string = self.spec_file.extract_version_from_archive_name(self.conf.sources,
                                                                              self.spec_file.get_main_source())
        else:
            logger.verbose("argument passed as a new source is a version")
            version_string = self.conf.sources
        version, extra_version = SpecFile.split_version_string(version_string, self.spec_file.header.version)
        self.rebase_spec_file.set_version(version)
        self.rebase_spec_file.set_extra_version(extra_version, version != self.spec_file.header.version)

        oldver = parse_version(self.spec_file.header.version)
        newver = parse_version(self.rebase_spec_file.header.version)
        oldex = self.spec_file.parse_release()[2]
        newex = extra_version

        if not self.conf.skip_version_check and (newver < oldver or (newver == oldver and newex == oldex)):
            raise RebaseHelperError("Current version is equal to or newer than the requested version, nothing to do.")

        self.rebase_spec_file.update_changelog(self.conf.changelog_entry)

        # run spec hooks
        plugin_manager.spec_hooks.run(self.spec_file, self.rebase_spec_file, **self.kwargs)

        # spec file object has been sanitized downloading can proceed
        if not self.conf.not_download_sources:
            for spec_file in [self.spec_file, self.rebase_spec_file]:
                spec_file.download_remote_sources()
                # parse spec again with sources downloaded to properly expand %prep section
                spec_file.update()
        # all local sources have been downloaded; we can check for name changes
        self._sanitize_sources()

    def _sanitize_sources(self) -> None:
        """Renames local sources whose name changed after version bump.

        For example if the specfile contains a Patch such as %{name}-%{version}.patch,
        the name changes after changing the version and the rebase would fail due to
        a missing patch whilst building SRPM.

        This method tries to correct such cases and rename the local file to match
        the new name. Modifies the rebase_spec_file object to contain the correct
        paths.
        """
        for source, index, source_type in self.rebase_spec_file.spc.sources:
            if urllib.parse.urlparse(source).scheme or source_type == 0:
                continue
            if os.path.exists(os.path.join(self.execution_dir, source)):
                continue
            # Find matching source in the old spec
            sources = [n for n, i, t in self.spec_file.spc.sources if i == index and t == source_type]
            if not sources:
                logger.error('Failed to find the source corresponding to %s in old version spec', source)
                continue
            source_old = sources[0]

            # rename the source
            old_source_path = os.path.join(self.rebased_sources_dir, source_old)
            new_source_path = os.path.join(self.rebased_sources_dir, source)
            logger.debug('Renaming %s to %s', old_source_path, new_source_path)
            try:
                os.rename(old_source_path, new_source_path)
            except OSError:
                logger.error('Failed to rename %s to %s while sanitizing sources', old_source_path, new_source_path)

            # prepend the Source/Path with rebased-sources directory in the specfile
            to_prepend = os.path.relpath(self.rebased_sources_dir, self.execution_dir)
            tag = '{0}{1}'.format('Patch' if source_type == 2 else 'Source', index)
            value = self.rebase_spec_file.get_raw_tag_value(tag)
            self.rebase_spec_file.set_raw_tag_value(tag, os.path.join(to_prepend, value))
        self.rebase_spec_file.save()

    def _initialize_data(self):
        """Function fill dictionary with default data"""
        # Get all tarballs before self.kwargs initialization
        self.old_sources = self.spec_file.get_archive()
        new_sources = self.rebase_spec_file.get_archive()

        self.old_sources = os.path.abspath(self.old_sources)
        if new_sources:
            self.conf.sources = new_sources

        if not self.conf.sources:
            raise RebaseHelperError('You have to define new sources.')
        else:
            self.new_sources = os.path.abspath(self.conf.sources)
        # Contains all sources except the main source
        self.old_rest_sources = [os.path.abspath(x) for x in self.spec_file.get_sources()[1:]]
        self.new_rest_sources = [os.path.abspath(x) for x in self.rebase_spec_file.get_sources()[1:]]

    def _find_spec_file(self) -> str:
        """Finds a spec file in the execution_dir directory.

        Returns:
            Path to the spec file.

        Raises:
            RebaseHelperError: If no spec file could be found.

        """
        spec_file_path = PathHelper.find_first_file(self.execution_dir, '*.spec', 0)
        if not spec_file_path:
            raise RebaseHelperError("Could not find any SPEC file "
                                    "in the current directory '{}'".format(self.execution_dir))
        return spec_file_path

    def _delete_old_builds(self):
        """
        Deletes the old and new result dir from previous build

        :return:
        """
        self._delete_new_results_dir()
        self._delete_old_results_dir()

    def _delete_old_results_dir(self):
        """
        Deletes old result dir

        :return:
        """
        if os.path.isdir(os.path.join(self.results_dir, constants.OLD_BUILD_DIR)):
            shutil.rmtree(os.path.join(self.results_dir, constants.OLD_BUILD_DIR))

    def _delete_new_results_dir(self):
        """
        Deletes new result dir

        :return:
        """
        if os.path.isdir(os.path.join(self.results_dir, constants.NEW_BUILD_DIR)):
            shutil.rmtree(os.path.join(self.results_dir, constants.NEW_BUILD_DIR))

    def _delete_workspace_dir(self):
        """
        Deletes workspace directory and loggs message

        :return:
        """
        logger.verbose("Removing the workspace directory '%s'", self.workspace_dir)
        if os.path.isdir(self.workspace_dir):
            shutil.rmtree(self.workspace_dir)

    def _check_workspace_dir(self):
        """
        Check if workspace dir exists, and removes it if yes.

        :return:
        """
        if os.path.exists(self.workspace_dir):
            logger.warning("Workspace directory '%s' exists, removing it", os.path.basename(self.workspace_dir))
            self._delete_workspace_dir()
        os.makedirs(self.workspace_dir)

    @staticmethod
    def _check_results_dir(results_dir):
        """
        Check if  results dir exists, and removes it if yes.

        :return:
        """
        # TODO: We may not want to delete the directory in the future
        if os.path.exists(results_dir):
            logger.warning("Results directory '%s' exists, removing it", os.path.basename(results_dir))
            shutil.rmtree(results_dir)
        os.makedirs(results_dir)
        os.makedirs(os.path.join(results_dir, constants.OLD_BUILD_DIR))
        os.makedirs(os.path.join(results_dir, constants.NEW_BUILD_DIR))
        os.makedirs(os.path.join(results_dir, constants.CHECKERS_DIR))
        os.makedirs(os.path.join(results_dir, constants.REBASED_SOURCES_DIR))

    @staticmethod
    def extract_archive(archive_path, destination):
        """
        Extracts given archive into the destination and handle all exceptions.

        :param archive_path: path to the archive to be extracted
        :param destination: path to a destination, where the archive should be extracted to
        :return:
        """
        archive = Archive(archive_path)

        try:
            archive.extract_archive(destination)
        except IOError as e:
            raise RebaseHelperError("Archive '{}' can not be extracted".format(archive_path)) from e
        except (EOFError, SystemError) as e:
            raise RebaseHelperError("Archive '{}' is damaged".format(archive_path)) from e

    @staticmethod
    def extract_sources(archive_path, destination):
        """Function extracts a given Archive and returns a full dirname to sources"""
        try:
            Application.extract_archive(archive_path, destination)
        except NotImplementedError:
            # not a standard archive type, can't extract it, fallback to copying
            os.makedirs(destination)
            shutil.copy(archive_path, destination)

        files = os.listdir(destination)

        if not files:
            raise RebaseHelperError('Extraction of sources failed!')
        # if there is only one directory, we can assume it's top-level directory
        elif len(files) == 1:
            sources_dir = os.path.join(destination, files[0])
            if os.path.isdir(sources_dir):
                return sources_dir

        # archive without top-level directory
        return destination

    def prepare_sources(self):
        """
        Function prepares a sources.

        :return:
        """

        old_sources_dir = os.path.join(self.workspace_dir, constants.OLD_SOURCES_DIR)
        new_sources_dir = os.path.join(self.workspace_dir, constants.NEW_SOURCES_DIR)

        old_dir = Application.extract_sources(self.old_sources, old_sources_dir)
        new_dir = Application.extract_sources(self.new_sources, new_sources_dir)

        old_tld = os.path.relpath(old_dir, old_sources_dir)
        new_tld = os.path.relpath(new_dir, new_sources_dir)

        dirname = self.spec_file.get_setup_dirname()

        if dirname and os.sep in dirname:
            dirs = os.path.split(dirname)
            if old_tld == dirs[0]:
                old_dir = os.path.join(old_dir, *dirs[1:])
            if new_tld == dirs[0]:
                new_dir = os.path.join(new_dir, *dirs[1:])

        new_dirname = os.path.relpath(new_dir, new_sources_dir)

        if new_dirname != '.':
            self.rebase_spec_file.update_setup_dirname(new_dirname)

        # extract rest of source archives to correct paths
        rest_sources = [self.old_rest_sources, self.new_rest_sources]
        spec_files = [self.spec_file, self.rebase_spec_file]
        sources_dirs = [old_sources_dir, new_sources_dir]
        for sources, spec_file, sources_dir in zip(rest_sources, spec_files, sources_dirs):
            for rest in sources:
                archive = [x for x in Archive.get_supported_archives() if rest.endswith(x)]
                if archive:
                    dest_dir = spec_file.find_archive_target_in_prep(rest)
                    if dest_dir:
                        Application.extract_sources(rest, os.path.join(sources_dir, dest_dir))

        return [old_dir, new_dir]

    def patch_sources(self, sources):
        try:
            # Patch sources
            self.rebased_patches = Patcher.patch(sources[0],
                                                 sources[1],
                                                 self.old_rest_sources,
                                                 self.spec_file.get_applied_patches(),
                                                 **self.kwargs)
        except RuntimeError as e:
            raise RebaseHelperError('Patching failed') from e
        self.rebase_spec_file.write_updated_patches(self.rebased_patches,
                                                    self.conf.disable_inapplicable_patches)
        results_store.set_patches_results(self.rebased_patches)

    def generate_patch(self):
        """
        Generates patch to the results_dir containing all needed changes for
        the rebased package version
        """
        # Delete removed patches from rebased_sources_dir from git
        removed_patches = self.rebase_spec_file.removed_patches
        if removed_patches:
            self.rebased_repo.index.remove(removed_patches, working_tree=True)

        self.rebase_spec_file.update_paths_to_sources_and_patches()

        # Generate patch
        self.rebased_repo.git.add(all=True)
        self.rebase_spec_file.update()
        self.rebased_repo.index.commit(MacroHelper.expand(self.conf.changelog_entry, self.conf.changelog_entry))
        patch = self.rebased_repo.git.format_patch('-1', stdout=True, stdout_as_string=False)
        with open(os.path.join(self.results_dir, constants.CHANGES_PATCH), 'wb') as f:
            f.write(patch)
            f.write(b'\n')

        results_store.set_changes_patch('changes_patch', os.path.join(self.results_dir, constants.CHANGES_PATCH))

    @classmethod
    def _update_gitignore(cls, sources, rebased_sources_dir):
        """Adds new entries into .gitignore file.

        Args:
            sources (list): List of new source files.
            rebased_sources_dir (str): Target directory.

        """
        gitignore = os.path.join(rebased_sources_dir, '.gitignore')

        if not os.path.isfile(gitignore):
            return

        with open(gitignore) as f:
            entries = f.readlines()

        def match(source):
            source = source.lstrip(os.path.sep).rstrip('\n')
            for entry in entries:
                if fnmatch.fnmatch(source, entry.lstrip(os.path.sep).rstrip('\n')):
                    return True
            return False

        with open(gitignore, 'a') as f:
            for src in [s for s in sources if not match(s)]:
                f.write(os.path.sep + src + '\n')

    def _prepare_rebased_repository(self):
        """
        Initialize git repository in the rebased directory
        :return: git.Repo instance of rebased_sources
        """
        for source, _, source_type in self.spec_file.spc.sources:
            # copy only existing local sources
            if not urllib.parse.urlparse(source).scheme and source_type == 1:
                source_path = os.path.join(self.execution_dir, source)
                if os.path.isfile(source_path):
                    shutil.copy(source_path, self.rebased_sources_dir)

        for patch in self.spec_file.get_applied_patches() + self.spec_file.get_not_used_patches():
            shutil.copy(patch.path, self.rebased_sources_dir)

        sources = os.path.join(self.execution_dir, 'sources')
        if os.path.isfile(sources):
            shutil.copy(sources, self.rebased_sources_dir)

        gitignore = os.path.join(self.execution_dir, '.gitignore')
        if os.path.isfile(gitignore):
            shutil.copy(gitignore, self.rebased_sources_dir)

        repo = git.Repo.init(self.rebased_sources_dir)
        repo.git.config('user.name', GitHelper.get_user(), local=True)
        repo.git.config('user.email', GitHelper.get_email(), local=True)
        repo.git.add(all=True)
        repo.index.commit('Initial commit', skip_hooks=True)
        return repo

    @staticmethod
    def _sanitize_build_dict(build_dict):
        blacklist = [
            'builds_nowait',
            'build_tasks',
            'builder_options',
            'srpm_builder_options',
            'app_kwargs',
        ]
        return {k: v for k, v in build_dict.items() if k not in blacklist}

    def build_source_packages(self):
        try:
            builder = plugin_manager.srpm_build_tools.get_plugin(self.conf.srpm_buildtool)
        except NotImplementedError as e:
            raise RebaseHelperError('{}. Supported SRPM build tools are {}'.format(
                str(e), ', '.join(plugin_manager.srpm_build_tools.get_supported_plugins()))) from e

        for version in ['old', 'new']:
            koji_build_id = None
            results_dir = os.path.join(self.results_dir, '{}-build'.format(version), 'SRPM')
            spec = self.spec_file if version == 'old' else self.rebase_spec_file
            package_name = spec.header.name
            package_version = spec.header.version
            logger.info('Building source package for %s version %s', package_name, package_version)

            if version == 'old' and self.conf.get_old_build_from_koji:
                koji_build_id, ver = KojiHelper.get_old_build_info(package_name, package_version)
                if ver:
                    package_version = ver

            build_dict = dict(
                name=package_name,
                version=package_version,
                srpm_buildtool=self.conf.srpm_buildtool,
                srpm_builder_options=self.conf.srpm_builder_options,
                app_kwargs=self.kwargs)
            try:
                os.makedirs(results_dir)
                if koji_build_id:
                    session = KojiHelper.create_session()
                    build_dict['srpm'], build_dict['logs'] = KojiHelper.download_build(session,
                                                                                       koji_build_id,
                                                                                       results_dir,
                                                                                       arches=['src'])

                else:
                    build_dict.update(builder.build(spec, results_dir, **build_dict))
                build_dict = self._sanitize_build_dict(build_dict)
                results_store.set_build_data(version, build_dict)
            except RebaseHelperError:  # pylint: disable=try-except-raise
                raise
            except SourcePackageBuildError as e:
                build_dict['logs'] = e.logs
                build_dict['source_package_build_error'] = str(e)
                build_dict = self._sanitize_build_dict(build_dict)
                results_store.set_build_data(version, build_dict)
                if e.logfile:
                    msg = 'Building {} SRPM packages failed; see {} for more information'.format(version, e.logfile)
                else:
                    msg = 'Building {} SRPM packages failed; see logs in {} for more information'.format(version,
                                                                                                         results_dir)
                raise RebaseHelperError(msg, logfiles=e.logs) from e
            except Exception as e:
                raise RebaseHelperError('Building package failed with unknown reason. '
                                        'Check all available log files.') from e

    def build_binary_packages(self):
        """Function calls build class for building packages"""
        try:
            builder = plugin_manager.build_tools.get_plugin(self.conf.buildtool)
        except NotImplementedError as e:
            raise RebaseHelperError('{}. Supported build tools are {}'.format(
                str(e), ', '.join(plugin_manager.build_tools.get_supported_plugins()))) from e

        for version in ['old', 'new']:
            results_dir = os.path.join(self.results_dir, '{}-build'.format(version), 'RPM')
            spec = None
            task_id = None
            koji_build_id = None
            build_dict: Dict[str, Any] = {}

            if self.conf.build_tasks is None:
                spec = self.spec_file if version == 'old' else self.rebase_spec_file
                package_name = spec.header.name
                package_version = spec.header.version

                if version == 'old' and self.conf.get_old_build_from_koji:
                    koji_build_id, ver = KojiHelper.get_old_build_info(package_name, package_version)
                    if ver:
                        package_version = ver

                build_dict = dict(
                    name=package_name,
                    version=package_version,
                    builds_nowait=self.conf.builds_nowait,
                    build_tasks=self.conf.build_tasks,
                    builder_options=self.conf.builder_options,
                    srpm=results_store.get_build(version).get('srpm'),
                    srpm_logs=results_store.get_build(version).get('logs'),
                    app_kwargs=self.kwargs)

                # prepare for building
                builder.prepare(spec, self.conf)

                logger.info('Building binary packages for %s version %s', package_name, package_version)
            else:
                task_id = self.conf.build_tasks[0] if version == 'old' else self.conf.build_tasks[1]

            try:
                os.makedirs(results_dir)
                if self.conf.build_tasks is None:
                    if koji_build_id:
                        session = KojiHelper.create_session()
                        build_dict['rpm'], build_dict['logs'] = KojiHelper.download_build(session,
                                                                                          koji_build_id,
                                                                                          results_dir,
                                                                                          arches=['noarch', 'x86_64'])
                    else:
                        build_dict.update(builder.build(spec, results_dir, **build_dict))
                if builder.CREATES_TASKS and task_id and not koji_build_id:
                    if not self.conf.builds_nowait:
                        build_dict['rpm'], build_dict['logs'] = builder.wait_for_task(build_dict,
                                                                                      task_id,
                                                                                      results_dir)
                    elif self.conf.build_tasks:
                        build_dict['rpm'], build_dict['logs'] = builder.get_detached_task(task_id, results_dir)
                build_dict = self._sanitize_build_dict(build_dict)
                results_store.set_build_data(version, build_dict)
            except RebaseHelperError:  # pylint: disable=try-except-raise
                # Proper RebaseHelperError instance was created already. Re-raise it.
                raise
            except BinaryPackageBuildError as e:
                build_dict['logs'] = e.logs
                build_dict['binary_package_build_error'] = str(e)
                build_dict = self._sanitize_build_dict(build_dict)
                results_store.set_build_data(version, build_dict)

                if e.logfile is None:
                    msg = 'Building {} RPM packages failed; see logs in {} for more information'.format(version,
                                                                                                        results_dir)
                else:
                    msg = 'Building {} RPM packages failed; see {} for more information'.format(version, e.logfile)

                raise RebaseHelperError(msg, logfiles=e.logs) from e
            except Exception as e:
                raise RebaseHelperError('Building package failed with unknown reason. '
                                        'Check all available log files.') from e

        if self.conf.builds_nowait and not self.conf.build_tasks:
            if builder.CREATES_TASKS:
                self.print_task_info(builder)

    def run_package_checkers(self, results_dir, **kwargs):
        """
        Runs checkers on packages and stores results in a given directory.

        :param results_dir: Path to directory in which to store the results.
        :type results_dir: str
        :param category: checker type(SOURCE/SRPM/RPM)
        :type category: str
        :return: None
        """
        results = dict()

        for checker_name in self.conf.pkgcomparetool:
            try:
                data = plugin_manager.checkers.run(os.path.join(results_dir, constants.CHECKERS_DIR),
                                                   checker_name,
                                                   **kwargs)
                if data:
                    results[checker_name] = data
            except CheckerNotFoundError:
                logger.error("Rebase-helper did not find checker '%s'.", checker_name)

        for diff_name, result in results.items():
            results_store.set_checker_output(diff_name, result)

    def get_new_build_logs(self):
        result: Dict[str, Dict[str, Dict[str, Any]]] = {}
        result['build_ref'] = {}
        for version in ['old', 'new']:
            result['build_ref'][version] = results_store.get_build(version)
        return result

    def print_summary(self, exception=None):
        """
        Save rebase-helper result and print the summary using output tools.
        :param exception: Error message from rebase-helper
        :return:
        """
        logs = None
        # Store rebase helper result exception
        if exception:
            if exception.logfiles:
                logs = exception.logfiles

            results_store.set_result_message('fail', exception.msg)
        else:
            result = "Rebase from {} to {} completed without an error".format(self.spec_file.get_NVR(),
                                                                              self.rebase_spec_file.get_NVR())
            results_store.set_result_message('success', result)

        if self.rebase_spec_file:
            self.rebase_spec_file.update_paths_to_sources_and_patches()
            self.generate_patch()

        plugin_manager.output_tools.run(self.conf.outputtool, logs, self)

    def print_task_info(self, builder):
        logs = self.get_new_build_logs()['build_ref']
        for version in ['old', 'new']:
            logger.info(builder.get_task_info(logs[version]))

    def apply_changes(self):
        try:
            repo = git.Repo(self.execution_dir)
        except git.InvalidGitRepositoryError:
            repo = git.Repo.init(self.execution_dir)
        patch = results_store.get_changes_patch()
        if not patch:
            logger.warning('Cannot apply %s. No patch file was created', constants.CHANGES_PATCH)
        try:
            repo.git.am(patch['changes_patch'])
        except git.GitCommandError as e:
            logger.warning('%s was not applied properly. Please review changes manually.'
                           '\nThe error message is: %s', constants.CHANGES_PATCH, str(e))

    def prepare_next_run(self, results_dir):
        # Running build log hooks only makes sense after a failed build
        # of new RPM packages. The folder results_dir/new-build/RPM
        # doesn't exist unless the build of new RPM packages has been run.
        changes_made = False
        if os.path.exists(os.path.join(results_dir, constants.NEW_BUILD_DIR, 'RPM')):
            changes_made = plugin_manager.build_log_hooks.run(self.spec_file, self.rebase_spec_file, **self.kwargs)
        # Save current rebase spec file content
        self.rebase_spec_file.save()
        if not self.conf.non_interactive and \
                InputHelper.get_message('Do you want to try it one more time'):
            logger.info('Now it is time to make changes to  %s if necessary.', self.rebase_spec_file.path)
        elif self.conf.non_interactive and changes_made:
            logger.info('Build log hooks made some changes to the SPEC file, starting the build process again.')
        else:
            return False
        if not self.conf.non_interactive and not \
                InputHelper.get_message('Do you want to continue with the rebuild now'):
            return False
        # Update rebase spec file content after potential manual modifications
        self.rebase_spec_file.reload()
        # clear current version output directories
        if os.path.exists(os.path.join(results_dir, constants.OLD_BUILD_DIR)):
            shutil.rmtree(os.path.join(results_dir, constants.OLD_BUILD_DIR))
        if os.path.exists(os.path.join(results_dir, constants.NEW_BUILD_DIR)):
            shutil.rmtree(os.path.join(results_dir, constants.NEW_BUILD_DIR))
        return True

    def run(self):
        # Certain options can be used only with specific build tools
        tools_creating_tasks = []
        for tool_name, tool in plugin_manager.build_tools.plugins.items():
            if tool and tool.CREATES_TASKS:
                tools_creating_tasks.append(tool_name)
        if self.conf.buildtool not in tools_creating_tasks:
            options_used = []
            if self.conf.build_tasks is not None:
                options_used.append('--build-tasks')
            if self.conf.builds_nowait is True:
                options_used.append('--builds-nowait')
            if options_used:
                raise RebaseHelperError("{} can be used only with the following build tools: {}".format(
                                        " and ".join(options_used),
                                        ", ".join(tools_creating_tasks)))
        elif self.conf.builds_nowait and self.conf.get_old_build_from_koji:
            raise RebaseHelperError("{} can't be used with: {}".format('--builds-nowait', '--get-old-build-from-koji'))

        tools_accepting_options = []
        for tool_name, tool in plugin_manager.build_tools.plugins.items():
            if tool and tool.ACCEPTS_OPTIONS:
                tools_accepting_options.append(tool_name)
        if self.conf.buildtool not in tools_accepting_options:
            options_used = []
            if self.conf.builder_options is not None:
                options_used.append('--builder-options')
            if options_used:
                raise RebaseHelperError("{} can be used only with the following build tools: {}".format(
                                        " and ".join(options_used),
                                        ", ".join(tools_accepting_options)))

        if self.conf.build_tasks is None:
            old_sources, new_sources = self.prepare_sources()
            self.run_package_checkers(self.results_dir,
                                      category=CheckerCategory.SOURCE,
                                      old_dir=old_sources,
                                      new_dir=new_sources)
            try:
                self.patch_sources([old_sources, new_sources])
            except RebaseHelperError as e:
                # Print summary and return error
                self.print_summary(e)
                raise

        # Build packages
        while True:
            try:
                if self.conf.build_tasks is None:
                    self.build_source_packages()
                self.run_package_checkers(self.results_dir, category=CheckerCategory.SRPM)
                self.build_binary_packages()
                if self.conf.builds_nowait and not self.conf.build_tasks:
                    return
                self.run_package_checkers(self.results_dir, category=CheckerCategory.RPM)
            # Print summary and return error
            except RebaseHelperError as e:
                logger.error(e.msg)
                if self.conf.build_tasks is None and self.prepare_next_run(self.results_dir):
                    continue
                self.print_summary(e)
                raise
            else:
                break

        if not self.conf.keep_workspace:
            self._delete_workspace_dir()

        self.print_summary()
        if self.conf.apply_changes:
            self.apply_changes()
        return 0



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import collections
import difflib
import fnmatch
import logging
import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.build_log_hooks import BaseBuildLogHook
from rebasehelper.types import PackageCategories
from rebasehelper.helpers.macro_helper import MacroHelper
from rebasehelper.constants import NEW_BUILD_DIR
from rebasehelper.specfile import SpecFile


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))

AddedFiles = Dict[str, List[str]]
RemovedFiles = Union[List[str], Dict[str, List[str]]]


class Files(BaseBuildLogHook):
    CATEGORIES: PackageCategories = []

    # taken from build/files.c in RPM source
    FILES_DIRECTIVES: Dict[str, Optional[str]] = {
        '%artifact': None,
        '%attr': None,
        '%caps': None,
        '%config': None,
        '%defattr': None,
        '%dev': None,
        '%dir': None,
        '%doc': '%{_defaultdocdir}',
        '%docdir': None,
        '%dverify': None,
        '%exclude': None,
        '%ghost': None,
        '%lang': None,
        '%license': '%{_defaultlicensedir}',
        '%missingok': None,
        '%pubkey': None,
        '%readme': None,
        '%verify': None,
    }

    # taken from build/rpmbuild_internal.h in rpm source code
    PROHIBITED_KEYWORDS: List[str] = [
        '%if',
        '%ifarch',
        '%ifnarch',
        '%ifos',
        '%ifnos',
        '%else',
        '%endif',
        '%elif',
        '%elifarch',
        '%elifos',
    ]

    @classmethod
    def format(cls, data):
        output = []
        for file_type, related_files in data.items():
            output.append(' - {}'.format(file_type))
            if isinstance(related_files, list):
                for file in related_files:
                    output.append('\t- {}'.format(file))
            else:
                for section, files in related_files.items():
                    output.append('\t- {}'.format(section))
                    for file in files:
                        output.append('\t\t- {}'.format(file))

        return output

    @classmethod
    def merge_two_results(cls, old, new):
        for file_type, related_files in new.items():
            if file_type not in old:
                old[file_type] = related_files
                continue

            if isinstance(related_files, list):
                old[file_type].extend(related_files)
                continue

            for section, files in related_files.items():
                if section not in old[file_type]:
                    old[file_type][section] = files
                    continue
                old[file_type][section].extend(files)

        return old

    @classmethod
    def _parse_build_log(cls, log_path, nvr):
        """Parses a build log.

        Args:
            log_path (str): Path to the RPM build log.

        Returns:
            tuple: The first element is the type of error that was found
                in the log (missing or deleted). The second element is
                a list of problematic files.

        """
        try:
            with open(log_path, 'r') as build_log:
                lines = build_log.read().splitlines()
        except IOError:
            logger.error('There was an error opening %s', log_path)
            return None, None

        error_type = None
        # Use set to avoid duplicate files.
        files = set()

        # Two types of error can occur. The file is either in the SPEC file but missing in sources
        # or the other way around. Example lines of the first case:
        # File not found: /root/rpmbuild/BUILDROOT/pello-0.1.1-1.fc27.x86_64/should/not/be/here
        # File not found: /root/rpmbuild/BUILDROOT/pello-0.1.1-1.fc27.x86_64/should/not/be/here2
        # Example lines of the second case:
        # Installed (but unpackaged) file(s) found:
        #   /usr/bin/pello
        error_re = re.compile(
            r'''
            ^
            (BUILDSTDERR:)?
            \s*
            (
                (?P<missing>File\s+not\s+found:\s*)|
                (?P<unpackaged>Installed\s+\(but\s+unpackaged\)\s+file\(s\)\s+found:)
            )?
            (/.*/{}\.\w+)?
            (?P<path>/.*)?
            $
            '''.format(re.escape(nvr)), re.VERBOSE
        )

        for line in lines:
            match = error_re.match(line)
            if match:
                if match.group('missing'):
                    error_type = 'deleted'
                    files.add(match.group('path'))
                elif match.group('unpackaged'):
                    error_type = 'missing'
                elif error_type == 'missing' and match.group('path'):
                    # Ignore debug information
                    if not match.group('path').startswith('/usr/lib/debug'):
                        files.add(match.group('path'))
                elif error_type and not match.group('path'):
                    break

        return error_type, list(files)

    @classmethod
    def _get_best_matching_files_section(cls, rebase_spec_file, file):
        """Finds a %files section with a file that has the closest match with
        the specified file. If the best match cannot be determined, the main
        %files section is returned. If no main section is found, return the
        first %files section if possible, None otherwise.

        Args:
            rebase_spec_file (specfile.SpecFile): Rebased SpecFile object.
            file (str): Path to the file to be classified.

        Returns:
            str: Name of the section containing the closest matching file.
                None if no %files section can be found.

        """
        best_match = ''
        best_match_section = ''
        files = []
        for sec_name, sec_content in rebase_spec_file.spec_content.sections:
            if sec_name.startswith('%files'):
                files.append(sec_name)
                for line in sec_content:
                    new_best_match = difflib.get_close_matches(file, [best_match, MacroHelper.expand(line)])
                    if new_best_match:
                        # the new match is a closer match
                        if new_best_match[0] != best_match:
                            best_match = str(new_best_match[0])
                            best_match_section = sec_name

        return best_match_section or rebase_spec_file.get_main_files_section() or (files[0] if files else None)

    @classmethod
    def _sanitize_path(cls, path):
        """Changes the path to follow Fedora Packaging Guidelines."""
        if path.startswith('%{_mandir}'):
            # substitute compression extension with *
            directory, name = os.path.split(path)
            name = '{0}.{1}*'.format(*name.split('.')[:2])
            path = os.path.join(directory, name)
        return path

    @classmethod
    def _correct_missing_files(cls, rebase_spec_file, files):
        """Adds files found in buildroot which are missing in %files
        sections in the SPEC file. Each file is added to a %files section
        with the closest matching path.

        """
        # get %{name} macro
        macros = [m for m in MacroHelper.filter(rebase_spec_file.macros, level=-3) if m['name'] == 'name']
        macros.extend(m for m in rebase_spec_file.macros if m['name'] in MacroHelper.MACROS_WHITELIST)
        macros = MacroHelper.expand_macros(macros)
        # ensure maximal greediness
        macros.sort(key=lambda k: len(k['value']), reverse=True)

        result: Dict[str, AddedFiles] = collections.defaultdict(lambda: collections.defaultdict(list))
        for file in files:
            section = cls._get_best_matching_files_section(rebase_spec_file, file)
            if section is None:
                logger.error('The specfile does not contain any %files section, cannot add the missing files')
                break
            substituted_path = cls._sanitize_path(MacroHelper.substitute_path_with_macros(file, macros))
            try:
                index = [i for i, l in enumerate(rebase_spec_file.spec_content.section(section)) if l][-1] + 1
            except IndexError:
                # section is empty
                index = 0
            rebase_spec_file.spec_content.section(section).insert(index, substituted_path)
            result['added'][section].append(substituted_path)
            logger.info("Added %s to '%s' section", substituted_path, section)

        return result

    @classmethod
    def _correct_deleted_files(cls, rebase_spec_file, files):
        """Removes files newly missing in buildroot from %files sections
        of the SPEC file. If a file cannot be removed, the user is informed
        and it is mentioned in the final report.

        """
        result: Dict[str, RemovedFiles] = collections.defaultdict(lambda: collections.defaultdict(list))
        for sec_name, sec_content in rebase_spec_file.spec_content.sections:
            if sec_name.startswith('%files'):
                subpackage = rebase_spec_file.get_subpackage_name(sec_name)
                i = 0
                while i < len(sec_content):
                    original_line = sec_content[i].split()
                    # Expand the whole line to check for occurrences of
                    # special keywords, such as %global and %if blocks.
                    # Macro definitions expand to empty string.
                    expanded = MacroHelper.expand(sec_content[i])
                    if not original_line or not expanded or any(k in expanded for k in cls.PROHIBITED_KEYWORDS):
                        i += 1
                        continue
                    split_line = original_line[:]
                    directives: List[str] = []
                    prepend_macro = None
                    for element in reversed(split_line):
                        if element in cls.FILES_DIRECTIVES:
                            if cls.FILES_DIRECTIVES[element]:
                                prepend_macro = cls.FILES_DIRECTIVES[element]
                            directives.insert(0, element)
                            split_line.remove(element)

                    if prepend_macro:
                        for j, path in enumerate(split_line):
                            if not os.path.isabs(path):
                                split_line[j] = os.path.join(prepend_macro, subpackage, os.path.basename(path))
                    split_line = [MacroHelper.expand(p) for p in split_line]

                    j = 0
                    while j < len(split_line) and files:
                        file = split_line[j]
                        for deleted_file in reversed(files):
                            if not fnmatch.fnmatch(deleted_file, file):
                                continue

                            original_file = original_line[len(directives) + j]

                            del split_line[j]
                            del original_line[len(directives) + j]
                            files.remove(deleted_file)
                            result['removed'][sec_name].append(original_file)
                            logger.info("Removed %s from '%s' section", original_file, sec_name)
                            break
                        else:
                            j += 1

                    if not split_line:
                        del sec_content[i]
                    else:
                        sec_content[i] = ' '.join(original_line)
                        i += 1

                    if not files:
                        return result

        logger.info('Could not remove the following files:')
        for file in files:
            logger.info('\t%s', file)

        result['unable_to_remove'] = files
        return result

    @classmethod
    def run(cls, spec_file: SpecFile, rebase_spec_file: SpecFile, results_dir: str,
            **kwargs: Any) -> Tuple[Dict[str, Union[RemovedFiles, AddedFiles]], bool]:
        if not results_dir:
            return {}, False
        log = os.path.join(results_dir, NEW_BUILD_DIR, 'RPM', 'build.log')

        nvr = rebase_spec_file.get_NVR()
        error_type, files = cls._parse_build_log(log, nvr)

        result: Dict[str, Union[AddedFiles, RemovedFiles]] = {}
        if error_type == 'deleted':
            logger.info('The following files are absent in sources but are in the SPEC file, trying to remove them:')
            for file in files:
                logger.info('\t%s', file)
            result = cls._correct_deleted_files(rebase_spec_file, files)
        elif error_type == 'missing':
            logger.info('The following files are in the sources but are missing in the SPEC file, trying to add them:')
            for file in files:
                logger.info('\t%s', file)
            result = cls._correct_missing_files(rebase_spec_file, files)
        rebase_spec_file.save()
        return result, 'added' in result or 'removed' in result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore

from typing import Dict, List, Union

from rebasehelper.results_store import ResultsStore


class TestResultsStore:
    old_rpm_data: Dict[str, Union[str, List[str]]] = {
        'rpm': ['rpm-0.1.0.x86_64.rpm', ' rpm-devel-0.1.0.x86_64.rpm'],
        'srpm': 'rpm-0.1.0.src.rpm',
        'logs': ['logfile1.log', 'logfile2.log']
    }
    new_rpm_data: Dict[str, Union[str, List[str]]] = {
        'rpm': ['rpm-0.2.0.x86_64.rpm', ' rpm-devel-0.2.0.x86_64.rpm'],
        'srpm': 'rpm-0.2.0.src.rpm',
        'logs': ['logfile3.log', 'logfile4.log']
    }
    patches_data: Dict[str, List[str]] = {
        'deleted': ['del_patch1.patch', 'del_patch2.patch'],
        'modified': ['mod_patch1.patch', 'mod_patch2.patch']
    }
    info_data: Dict[str, str] = {'Information text': 'some information text'}
    info_data2: Dict[str, str] = {'Next Information': 'some another information text'}

    @pytest.fixture
    def results_store(self):
        rs = ResultsStore()
        rs.set_info_text('Information text', 'some information text')
        rs.set_info_text('Next Information', 'some another information text')
        rs.set_patches_results(self.patches_data)
        rs.set_build_data('old', self.old_rpm_data)
        rs.set_build_data('new', self.new_rpm_data)
        return rs

    def test_base_output_info(self, results_store):
        """
        Test Output logger info

        :return:
        """
        info_results = results_store.get_summary_info()
        expect_dict = self.info_data
        expect_dict.update(self.info_data2)
        assert info_results == expect_dict

    def test_base_output_patches(self, results_store):
        """
        Test Output logger patches

        :return:
        """
        patch_results = results_store.get_patches()
        expected_patches = self.patches_data
        assert patch_results == expected_patches

    def test_base_output_builds_old(self, results_store):
        """
        Test Output logger old builds

        :return:
        """
        build_results = results_store.get_build('old')
        assert build_results == self.old_rpm_data

    def test_base_output_builds_new(self, results_store):
        """
        Test Output logger new builds

        :return:
        """
        build_results = results_store.get_build('new')
        assert build_results == self.new_rpm_data



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
import urllib.parse
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.temporary_environment import TemporaryEnvironment
from rebasehelper.types import PackageCategories
from rebasehelper.specfile import PackageCategory
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.macro_helper import MacroHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RubyHelper(BaseSpecHook):

    CATEGORIES: PackageCategories = [PackageCategory.ruby]

    @classmethod
    def _get_instructions(cls, comments, old_version, new_version):
        """Extract instructions from comments, update version if necessary"""
        instructions = []
        for comment in comments:
            comment = MacroHelper.expand(comment, comment)
            comment = MacroHelper.expand(comment, comment)
            comment = re.sub(r'^#\s*', '', comment)
            comment = comment.replace(old_version, new_version)
            instructions.append(comment)
        return instructions

    @classmethod
    def _build_source_from_instructions(cls, instructions, source, logfile):
        """Run instructions to create source archive"""
        logger.info("Attempting to create source '%s' using instructions in comments", source)
        with TemporaryEnvironment() as tmp:
            script = os.path.join(tmp.path(), 'script.sh')
            with open(script, 'w') as f:
                f.write('#!/bin/sh -x\n')
                f.write('{}\n'.format('\n'.join(instructions)))
                f.write('cp "{}" "{}"\n'.format(source, os.getcwd()))
            os.chmod(script, 0o755)
            result = ProcessHelper.run_subprocess_cwd(script, tmp.path(), output_file=logfile, shell=True)
        if result == 0 and os.path.isfile(source):
            logger.info('Source creation succeeded.')
        else:
            logger.info('Source creation failed.')

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        # find non-existent local sources
        sources = [idx for idx, src in enumerate(rebase_spec_file.sources)
                   if not urllib.parse.urlparse(src).scheme and not os.path.isfile(src)]
        for idx in sources:
            if spec_file.sources[idx] == rebase_spec_file.sources[idx]:
                # skip sources that stayed unchanged
                continue
            tag = rebase_spec_file.tag('Source{0}'.format(idx))
            if not tag:
                continue
            source = rebase_spec_file.sources[idx]
            logger.info("Found non-existent source '%s'", source)
            comments = None
            preamble = rebase_spec_file.spec_content.section('%package')
            for i in range(tag.line - 1, 0, -1):
                if not preamble[i].startswith('#'):
                    comments = preamble[i+1:i]
                    break
            if not comments:
                # nothing to do
                continue
            # update data so that RPM macros are populated correctly
            rebase_spec_file.update()
            instructions = cls._get_instructions(comments,
                                                 spec_file.header.version,
                                                 rebase_spec_file.header.version)
            logfile = os.path.join(kwargs['workspace_dir'], '{}.log'.format(source))
            cls._build_source_from_instructions(instructions, source, logfile)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import configparser
import hashlib
import logging
import os
import re
import sys
import time
from typing import cast

import requests
import requests_gssapi  # type: ignore
from urllib3.fields import RequestField  # type: ignore
from urllib3.filepost import encode_multipart_formdata  # type: ignore

from rebasehelper.exceptions import LookasideCacheError, DownloadError
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.download_helper import DownloadHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class LookasideCacheHelper:

    """Class for downloading files from Fedora/RHEL lookaside cache"""

    rpkg_config_dir: str = '/etc/rpkg'

    @classmethod
    def _read_config(cls, tool):
        config = configparser.ConfigParser()
        config.read(os.path.join(cls.rpkg_config_dir, '{}.conf'.format(tool)))
        return dict(config.items(tool, raw=True))

    @classmethod
    def _read_sources(cls, basepath):
        line_re = re.compile(r'^(?P<hashtype>[^ ]+?) \((?P<filename>[^ )]+?)\) = (?P<hash>[^ ]+?)$')
        sources = []
        path = os.path.join(basepath, 'sources')
        if os.path.isfile(path):
            with open(path, 'r') as f:
                for line in f.readlines():
                    line = line.strip()
                    m = line_re.match(line)
                    if m is not None:
                        d = m.groupdict()
                    else:
                        # fall back to old format of sources file
                        hsh, filename = line.split()
                        d = dict(hash=hsh, filename=filename, hashtype='md5')
                    d['hashtype'] = d['hashtype'].lower()
                    sources.append(d)
        elif os.path.exists(path):
            logger.warning("\"sources\" is not a file, skipping parsing it")
        return sources

    @classmethod
    def _write_sources(cls, basepath, sources):
        path = os.path.join(basepath, 'sources')
        with open(path, 'w') as f:
            for source in sources:
                f.write('{0} ({1}) = {2}\n'.format(source['hashtype'].upper(), source['filename'], source['hash']))

    @classmethod
    def _hash(cls, filename, hashtype):
        try:
            chksum = hashlib.new(hashtype)
        except ValueError as e:
            raise LookasideCacheError('Unsupported hash type \'{}\''.format(hashtype)) from e
        with open(filename, 'rb') as f:
            chunk = f.read(8192)
            while chunk:
                chksum.update(chunk)
                chunk = f.read(8192)
        return chksum.hexdigest()

    @classmethod
    def _download_source(cls, tool, url, package, filename, hashtype, hsh, target=None):
        if target is None:
            target = os.path.basename(filename)
        if os.path.exists(target):
            if cls._hash(target, hashtype) == hsh:
                # nothing to do
                return
            else:
                os.unlink(target)
        if tool == 'fedpkg':
            url = '{0}/{1}/{2}/{3}/{4}/{2}'.format(url, package, filename, hashtype, hsh)
        else:
            url = '{0}/{1}/{2}/{3}/{2}'.format(url, package, filename, hsh)
        try:
            DownloadHelper.download_file(url, target)
        except DownloadError as e:
            raise LookasideCacheError(str(e)) from e

    @classmethod
    def download(cls, tool, basepath, package, target_dir=None):
        try:
            config = cls._read_config(tool)
            url = config['lookaside']
        except (configparser.Error, KeyError) as e:
            raise LookasideCacheError('Failed to read rpkg configuration') from e
        for source in cls._read_sources(basepath):
            target = os.path.join(target_dir, source['filename'])
            cls._download_source(tool, url, package, source['filename'], source['hashtype'], source['hash'], target)

    @classmethod
    def _upload_source(cls, url, package, source_dir, filename, hashtype, hsh, auth=requests_gssapi.HTTPSPNEGOAuth()):
        class ChunkedData:
            def __init__(self, check_only, chunksize=8192):
                self.check_only = check_only
                self.chunksize = chunksize
                self.start = time.time()
                self.uploaded = False
                fields = [
                    ('name', package),
                    ('{}sum'.format(hashtype), hsh),
                ]
                if check_only:
                    fields.append(('filename', filename))
                else:
                    with open(path, 'rb') as f:
                        rf = RequestField('file', f.read(), filename)
                        rf.make_multipart()
                        fields.append(rf)
                self.data, content_type = encode_multipart_formdata(fields)
                self.headers = {'Content-Type': content_type}

            def __iter__(self):
                if self.uploaded:
                    # ensure the progressbar is shown only once (HTTPSPNEGOAuth causes second request)
                    yield self.data
                else:
                    totalsize = len(self.data)
                    for offset in range(0, totalsize, self.chunksize):
                        transferred = min(offset + self.chunksize, totalsize)
                        if not self.check_only:
                            DownloadHelper.progress(totalsize, transferred, self.start)
                        yield self.data[offset:transferred]
                    self.uploaded = True

        def post(check_only=False):
            cd = ChunkedData(check_only)
            r = requests.post(url, data=cd, headers=cd.headers, auth=auth)
            if not 200 <= r.status_code < 300:
                raise LookasideCacheError(r.reason)
            return r.content

        path = os.path.join(source_dir, filename)
        state = post(check_only=True)
        if state.strip() == b'Available':
            # already uploaded
            return

        logger.info('Uploading %s to lookaside cache', path)
        try:
            post()
        except requests.exceptions.ConnectionError as e:
            # Skip error, the rebase can continue even after a failed upload
            logger.error("Upload to lookaside cache failed: %s", str(e))
        finally:
            sys.stdout.write('\n')
            sys.stdout.flush()

    @classmethod
    def update_sources(cls, tool, basepath, package, old_sources, new_sources, upload=True, source_dir=''):
        try:
            config = cls._read_config(tool)
            url = config['lookaside_cgi']
            hashtype = config['lookasidehash']
        except (configparser.Error, KeyError) as e:
            raise LookasideCacheError('Failed to read rpkg configuration') from e
        uploaded = []
        sources = cls._read_sources(basepath)
        for idx, src in enumerate(old_sources):
            indexes = [i for i, s in enumerate(sources) if s['filename'] == src]
            if indexes:
                filename = new_sources[idx]
                if filename == src:
                    # no change
                    continue
                hsh = cls._hash(filename, hashtype)
                if upload:
                    cls._upload_source(url, package, source_dir, filename, hashtype, hsh)
                uploaded.append(filename)
                sources[indexes[0]] = dict(hash=hsh, filename=filename, hashtype=hashtype)
        cls._write_sources(basepath, sources)
        return uploaded



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import rpm  # type: ignore

from rebasehelper.helpers.macro_helper import MacroHelper


class TestMacroHelper:

    def test_get_macros(self):
        rpm.addMacro('test_macro', 'test_macro value')
        macros = MacroHelper.dump()
        macros = MacroHelper.filter(macros, name='test_macro', level=-1)
        assert len(macros) == 1
        assert macros[0]['name'] == 'test_macro'
        assert macros[0]['value'] == 'test_macro value'
        assert macros[0]['level'] == -1



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io
import os
from typing import Optional

from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.exceptions import CheckerNotFoundError
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory


class CsMock(BaseChecker):
    """csmock checker"""

    CATEGORY: Optional[CheckerCategory] = CheckerCategory.SRPM

    CMD: str = 'csmock'

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 0
        except (IOError, OSError):
            return False

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """Compares old and new RPMs using pkgdiff"""
        csmock_report = {}

        old_pkgs = results_store.get_old_build().get('srpm', None)
        new_pkgs = results_store.get_new_build().get('srpm', None)
        cls.results_dir = os.path.join(results_dir, cls.CMD)
        cls.prepare_results_dir()
        arguments = ['--force', '-a', '-r', 'fedora-rawhide-x86_64', '--base-srpm']
        if old_pkgs and new_pkgs:
            cmd = [cls.CMD]
            cmd.extend(arguments)
            cmd.append(old_pkgs)
            cmd.append(new_pkgs)
            cmd.extend(['-o', results_dir])
            output = io.StringIO()
            try:
                ProcessHelper.run_subprocess(cmd, output_file=output)
            except OSError as e:
                raise CheckerNotFoundError("Checker '{}' was not found or installed.".format(cls.name)) from e
        csmock_report['error'] = PathHelper.find_all_files_current_dir(results_dir, '*.err')
        csmock_report['txt'] = PathHelper.find_all_files_current_dir(results_dir, '*.txt')
        csmock_report['log'] = PathHelper.find_all_files_current_dir(results_dir, '*.log')
        csmock_report['path'] = cls.get_checker_output_dir_short()
        return csmock_report

    @classmethod
    def format(cls, data):
        """
        Formats csmock data to string
        :param data: csmock data dictionary
        :return: string formated output
        """
        output_lines = [cls.get_underlined_title("csmock")]
        output_lines.append("Details in {}:".format(data['path']))
        for key, files_list in data.items():
            if key in ['error', 'txt', 'log'] and files_list:
                for f in files_list:
                    output_lines.append(" - {}".format(os.path.basename(f)))

        return output_lines



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
from typing import cast

import koji  # type: ignore  # pylint: disable=import-error

# unused import needed to prevent loading koji buildtool with Koji < 1.13
import koji_cli.lib  # type: ignore  # pylint: disable=import-error,unused-import

from rebasehelper.types import Options
from rebasehelper.helpers.koji_helper import KojiHelper
from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.plugins.build_tools.rpm import BuildToolBase
from rebasehelper.plugins.build_tools import get_mock_logfile_path
from rebasehelper.exceptions import BinaryPackageBuildError
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Koji(BuildToolBase):
    """
    Class representing Koji build tool.
    """

    OPTIONS: Options = [
        {
            "name": ["--get-old-build-from-koji"],
            "default": False,
            "switch": True,
            "help": "do not build old sources, download latest build from Koji instead",
        },
    ]

    CREATES_TASKS: bool = True
    TARGET_TAG: str = 'rawhide'

    @classmethod
    def _verify_tasks(cls, session, task_dict):
        """Checks if any of the tasks failed and tries to extract mock exit code from it.

        Args:
            session (koji.ClientSession): Active Koji session instance.
            task_dict (dict): Dict mapping Koji task ID to its state.

        Returns:
            int: Mock exit code or -1 if any task failed, otherwise None.

        """
        for task_id, state in task_dict.items():
            if state == koji.TASK_STATES['FAILED']:
                try:
                    session.getTaskResult(task_id)
                except koji.BuildError as e:
                    # typical error message:
                    #   BuildError: error building package (arch noarch),
                    #   mock exited with status 1; see build.log for more information
                    match = re.search(r'mock exited with status (\d+)', str(e))
                    if match:
                        return int(match.group(1))
                    else:
                        return -1
        return None

    @classmethod
    def _scratch_build(cls, results_dir, srpm, **kwargs):
        session = KojiHelper.create_session(login=True)
        remote = KojiHelper.upload_srpm(session, srpm)
        task_id = session.build(remote, cls.TARGET_TAG, dict(scratch=True))
        if kwargs['builds_nowait']:
            return None, None, task_id
        url = KojiHelper.get_task_url(session, task_id)
        logger.info('Koji task is here: %s\n', url)
        session.logout()
        task_dict = KojiHelper.watch_koji_tasks(session, [task_id])
        path = os.path.dirname(srpm).replace('SRPM', 'RPM')
        rpms, logs = KojiHelper.download_task_results(session, list(task_dict), path)
        exit_code = cls._verify_tasks(session, task_dict)
        if exit_code:
            raise BinaryPackageBuildError(logs=logs, logfile=get_mock_logfile_path(exit_code, results_dir))
        return rpms, logs, task_id

    @classmethod
    def wait_for_task(cls, build_dict, task_id, results_dir):
        session = KojiHelper.create_session()
        task_dict = KojiHelper.watch_koji_tasks(session, [task_id])
        rpms, logs = KojiHelper.download_task_results(session, list(task_dict), results_dir)
        exit_code = cls._verify_tasks(session, task_dict)
        if exit_code:
            raise BinaryPackageBuildError(logs=logs, logfile=get_mock_logfile_path(exit_code, results_dir))
        return rpms, logs

    @classmethod
    def get_task_info(cls, build_dict):
        session = KojiHelper.create_session()
        url = KojiHelper.get_task_url(session, build_dict['koji_task_id'])
        return 'Scratch build for {} version is: {}'.format(build_dict['version'], url)

    @classmethod
    def get_detached_task(cls, task_id, results_dir):
        session = KojiHelper.create_session()
        rpms, logs = KojiHelper.download_task_results(session, [task_id], results_dir)
        task = session.getTaskInfo(task_id)
        exit_code = cls._verify_tasks(session, {task_id: task['state']})
        if exit_code:
            raise BinaryPackageBuildError(logs=logs, logfile=get_mock_logfile_path(exit_code, results_dir))
        if not rpms:
            raise RebaseHelperError('Koji tasks are not finished yet. Try again later.')
        return rpms, logs

    @classmethod
    def build(cls, spec, results_dir, srpm, **kwargs):
        """
        Builds the RPMs using koji

        :param spec: SpecFile object
        :param results_dir: absolute path to DIR where results should be stored
        :param srpm: absolute path to SRPM
        :param upstream_monitoring: specify if build is handled by upstream monitoring
        :return: dict with:
                 'rpm' -> list with absolute paths to RPMs
                 'logs' -> list with absolute paths to build_logs
                 'koji_task_id' -> ID of koji task
        """
        rpms, logs, koji_task_id = cls._scratch_build(results_dir, srpm, **kwargs)
        return dict(rpm=rpms, logs=logs, koji_task_id=koji_task_id)



import hashlib
import hmac
import http.server
import json
import os
import sys
import tempfile
import threading
import traceback

import distutils.core
import git
import twine.commands.upload


class PyPI:

    @staticmethod
    def release(url, tag):
        with tempfile.TemporaryDirectory() as wd:
            os.chdir(wd)
            repo = git.Repo.clone_from(url, wd)
            repo.git.checkout(tag)
            sys.path.insert(0, wd)
            distutils.core.run_setup(os.path.join(wd, 'setup.py'),
                                     ['sdist', 'bdist_wheel', '--universal'])
            twine.commands.upload.main([os.path.join(wd, 'dist', '*')])


class Worker(threading.Thread):

    def __init__(self):
        super(Worker, self).__init__(daemon=True)
        self.condition = threading.Condition()
        self.url = None
        self.tag = None

    def run(self):
        while True:
            with self.condition:
                self.condition.wait()
                try:
                    PyPI.release(self.url, self.tag)
                except Exception as e:
                    traceback.print_exception(type(e), e, e.__traceback__)
                    continue


class HTTPServer:

    def __init__(self, port):
        class RequestHandler(http.server.BaseHTTPRequestHandler):

            def do_POST(self):
                def verify(signature, data):
                    try:
                        sha, signature = signature.split('=')
                    except ValueError:
                        return False
                    if sha != 'sha1':
                        return False
                    mac = hmac.new(os.environb.get(b'SECRET', bytes()), data, hashlib.sha1)
                    return hmac.compare_digest(mac.hexdigest(), signature)

                length = int(self.headers.get('Content-Length', 0))
                data = self.rfile.read(length)
                if not verify(self.headers.get('X-Hub-Signature'), data):
                    self.send_error(403)
                    return
                event = self.headers.get('X-GitHub-Event')
                if event == 'ping':
                    self.send_response(204)
                    self.end_headers()
                    return
                if event == 'release':
                    data = json.loads(data)
                    if self.server.worker.condition.acquire(blocking=False):
                        try:
                            self.server.worker.url = data.get('repository', {}).get('clone_url')
                            self.server.worker.tag = data.get('release', {}).get('tag_name')
                            self.server.worker.condition.notify()
                        finally:
                            self.server.worker.condition.release()
                self.send_response(204)
                self.end_headers()

        self.server = http.server.HTTPServer(('', port), RequestHandler)
        self.server.worker = Worker()

    def serve(self):
        self.server.worker.start()
        self.server.serve_forever()


def main():
    try:
        HTTPServer(80).serve()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# rebase-helper documentation build configuration file
#

import os
import sys

sys.path[0:0] = [os.path.abspath('.'), os.path.abspath('..')]

from rebasehelper import VERSION

# Determine if this script is running inside RTD build environment
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:
    # Use RTD theme when building locally
    import sphinx_rtd_theme


# -- General configuration ------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'ext.autoargs',
    'ext.custom_man_builder',
    'm2r',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = ['.rst', '.md']

# The encoding of source files.
source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'rebase-helper'
copyright = '2014-2019, Red Hat'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '.'.join(VERSION.split('.')[0:2])
# The full version, including alpha/beta/rc tags.
release = VERSION

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = []


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'default' if on_rtd else 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [] if on_rtd else [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'rebase-helper {0} documentation'.format(VERSION)

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = html_title

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom style sheets
html_css_files = ['css/custom.css']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'rebase-helper-docs'


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('man/rebase-helper', 'rebase-helper',
     'helps you rebase your package to the latest version',
     ['Petr Hráček <phracek@redhat.com>',
      'Tomáš Hozza <thozza@redhat.com>',
      'Nikola Forró <nforro@redhat.com>',
      'František Nečas <fifinecas@seznam.cz>'], 1),
]

# If true, show URL addresses after external links.
man_show_urls = False


# -- Options for autodoc --------------------------------------------------

# This value contains a list of modules to be mocked up. This is useful
# when some external dependencies are not met at build time and break
# the building process.
autodoc_mock_imports = ['rpm', 'koji', 'requests_gssapi']



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore

from rebasehelper.spec_content import SpecContent
from rebasehelper.tests.conftest import SPEC_FILE


class TestSpecContent:
    TEST_FILES = [
        SPEC_FILE
    ]

    @pytest.fixture
    def spec_content(self):
        with open(SPEC_FILE, 'r') as infile:
            return SpecContent(infile.read())

    def test_string_representation(self, spec_content):
        with open(SPEC_FILE, 'r') as infile:
            assert str(spec_content) == infile.read()

    @pytest.mark.parametrize('line, section, expected', [
        ('#test comment', '%package', (0, 13)),
        ('make install', '%install', (12, 12)),
        ('make install # install', '%install', (12, 22)),
        ('Name: test #invalid comment', '%package', (27, 27)),
    ], ids=[
        'whole_line',
        'no_comment',
        'inline_allowed',
        'inline_prohibited',
    ])
    def test_get_comment_span(self, line, section, expected):
        assert SpecContent.get_comment_span(line, section) == expected

    @pytest.mark.parametrize('section, expected', [
        ('%install', ['make DESTDIR=$RPM_BUILD_ROOT install', '']),
        ('%package test', None),

    ], ids=[
        'existent',
        'nonexistent',
    ])
    def test_section(self, spec_content, section, expected):
        assert spec_content.section(section) == expected

    def test_replace_section(self, spec_content):
        spec_content.replace_section('%install', ['#removed'])
        assert spec_content.section('%install') == ['#removed']



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import json
import os

import pytest  # type: ignore

from rebasehelper.plugins.output_tools.json_ import JSON
from rebasehelper.plugins.output_tools.text import Text
from rebasehelper.plugins.plugin_manager import plugin_manager
from rebasehelper.results_store import ResultsStore


class TestOutputTool:
    @pytest.fixture
    def results_file_path(self, workdir):
        return os.path.join(workdir, 'output_file')

    @pytest.fixture
    def results_store(self, workdir):
        data = {'old': {'patches_full': {0: ['mytest.patch', '-p1', 0],
                                         1: ['mytest2.patch', '-p1', 1]},
                        'srpm': './test-1.2.0-1.src.rpm',
                        'rpm': ['./test-1.2.0-1.x86_64.rpm', './test-devel-1.2.0-1.x86_64.rpm'],
                        'logs': ['logfile1.log', 'logfile2.log']
                        },
                'new': {'patches_full': {0: ['mytest.patch', 0, '-p1'],
                                         1: ['mytest2.patch', 1, '-p1']},
                        'srpm': './test-1.2.2-1.src.rpm',
                        'rpm': ['./test-1.2.2-1.x86_64.rpm', './test-devel-1.2.2-1.x86_64.rpm'],
                        'logs': ['logfile3.log', 'logfile4.log']},
                'patches': {'deleted': ['mytest2.patch']},
                'results_dir': workdir,
                'moved': ['/usr/sbin/test', '/usr/sbin/test2'],
                'success': 'Success',
                }

        rs = ResultsStore()
        rs.set_build_data('old', data['old'])
        rs.set_build_data('new', data['new'])
        rs.set_patches_results(data['patches'])
        rs.set_checker_output('pkgdiff', {'path': 'rebase-helper-results/checkers/pkgdiff'})
        files_build_log_hook_result = {
            'removed': {'%files': ['README']}
        }
        rs.set_build_log_hooks_result('files', files_build_log_hook_result)
        rs.set_info_text('Information text', 'some information text')
        rs.set_info_text('Next Information', 'some another information text')
        rs.set_result_message('success', data['success'])

        return rs

    def get_expected_text_output(self, workdir):
        expected_output = """\
Success
All result files are stored in {workdir}

pkgdiff
=======
Details in rebase-helper-results/checkers/pkgdiff:
- report.html
- report.txt

files build log hook
====================
- removed
- %files
- README

Downstream Patches
==================
Rebased patches are located in rebase-helper-results/rebased-sources
Legend:
[-] = already applied, patch removed
[*] = merged, patch modified
[!] = conflicting or inapplicable, patch skipped
[ ] = patch untouched
* mytest2.patch                            [-]

RPMS
====

Old packages
------------

Source packages and logs are in directory rebase-helper-results/old-build/SRPM:
- test-1.2.0-1.src.rpm
- logfile1.log
- logfile2.log

Binary packages and logs are in directory rebase-helper-results/old-build/RPM:
- test-1.2.0-1.x86_64.rpm
- test-devel-1.2.0-1.x86_64.rpm
- logfile1.log
- logfile2.log

New packages
------------

Source packages and logs are in directory rebase-helper-results/new-build/SRPM:
- test-1.2.2-1.src.rpm
- logfile3.log
- logfile4.log

Binary packages and logs are in directory rebase-helper-results/new-build/RPM:
- test-1.2.2-1.x86_64.rpm
- test-devel-1.2.2-1.x86_64.rpm
- logfile3.log
- logfile4.log""".format(workdir=workdir)
        return expected_output

    @staticmethod
    def get_expected_json_output():
        expected_output = {
            ResultsStore.RESULTS_BUILDS: {
                "new": {
                    "logs": ["logfile3.log", "logfile4.log"],
                    "patches_full": {
                        "0": ["mytest.patch", 0, "-p1"],
                        "1": ["mytest2.patch", 1, "-p1"]
                    },
                    "rpm": ["./test-1.2.2-1.x86_64.rpm", "./test-devel-1.2.2-1.x86_64.rpm"],
                    "srpm": "./test-1.2.2-1.src.rpm"
                },
                "old": {
                    "logs": ["logfile1.log", "logfile2.log"],
                    "patches_full": {
                        "0": ["mytest.patch", "-p1", 0],
                        "1": ["mytest2.patch", "-p1", 1]
                    },
                    "rpm": ["./test-1.2.0-1.x86_64.rpm", "./test-devel-1.2.0-1.x86_64.rpm"],
                    "srpm": "./test-1.2.0-1.src.rpm"
                }
            },
            ResultsStore.RESULTS_CHECKERS: {
                'pkgdiff': {
                    'path': 'rebase-helper-results/checkers/pkgdiff',
                },
            },
            ResultsStore.RESULTS_BUILD_LOG_HOOKS: {
                'files': {
                    'removed': {
                        '%files': ['README']
                    }
                }
            },
            ResultsStore.RESULTS_INFORMATION: {
                "Information text": "some information text",
                "Next Information": "some another information text"
            },
            ResultsStore.RESULTS_PATCHES: {
                "deleted": ["mytest2.patch"]
            },
            ResultsStore.RESULTS_SUCCESS: {
                "success": "Success"
            }
        }
        return expected_output

    def test_text_output_tool(self, results_file_path, results_store):
        assert Text.name in plugin_manager.output_tools.plugins
        Text.print_summary(results_file_path, results_store)

        with open(results_file_path) as f:
            lines = [y.strip() for y in f.readlines()]
            assert lines == self.get_expected_text_output(os.path.dirname(results_file_path)).split('\n')

    def test_json_output_tool(self, results_file_path, results_store):
        assert JSON.name in plugin_manager.output_tools.plugins
        JSON.print_summary(results_file_path, results_store)

        with open(results_file_path) as f:
            json_dict = json.load(f)
            # in Python2 strings in json decoded dict are Unicode, which would make the test fail
            assert json_dict == json.loads(json.dumps(self.get_expected_json_output()))



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re
from typing import List, Optional, Tuple


class SpecContent:
    """Class representing content of a SPEC file."""

    # taken from build/parseSpec.c in rpm source code
    SECTION_HEADERS: List[str] = [
        '%package',
        '%prep',
        '%generate_build_requires',
        '%build',
        '%install',
        '%check',
        '%clean',
        '%prerun',
        '%postrun',
        '%pretrans',
        '%posttrans',
        '%pre',
        '%post',
        '%files',
        '%changelog',
        '%description',
        '%triggerpostun',
        '%triggerprein',
        '%triggerun',
        '%triggerin',
        '%trigger',
        '%verifyscript',
        '%sepolicy',
        '%filetriggerin',
        '%filetrigger',
        '%filetriggerun',
        '%filetriggerpostun',
        '%transfiletriggerin',
        '%transfiletrigger',
        '%transfiletriggerun',
        '%transfiletriggerpostun',
        '%end',
        '%patchlist',
        '%sourcelist',
    ]

    # Comments in these sections can only be on a separate line.
    DISALLOW_INLINE_COMMENTS: List[str] = [
        '%package',
        '%patchlist',
        '%sourcelist',
        '%description',
        '%files',
        '%changelog',
    ]

    def __init__(self, content: str) -> None:
        self.sections: List[Tuple[str, List[str]]] = self._split_sections(content)

    def __str__(self) -> str:
        """Join SPEC file sections back together."""
        content = []
        for header, section in self.sections:
            if header != '%package':
                content.append(header + '\n')
            for line in section:
                content.append(line + '\n')
        return ''.join(content)

    def __getitem__(self, index: int) -> List[str]:
        return self.sections[index][1]

    @classmethod
    def get_comment_span(cls, line: str, section: str) -> Tuple[int, int]:
        """Gets span of a comment depending on the section.

        Args:
            line: Line to find the comment in.
            section: Section the line is in.

        Returns:
            Span of the comment. If no comment is found, both tuple elements
            are equal to the length of the line for convenient use in a slice.

        """
        inline_comment_allowed = not any(section.startswith(s) for s in cls.DISALLOW_INLINE_COMMENTS)
        comment = re.search(r" #.*" if inline_comment_allowed else r"^\s*#.*", line)
        return comment.span() if comment else (len(line), len(line))

    def section(self, name: str) -> Optional[List[str]]:
        """Gets content of a section.

        In case there are multiple sections with the same name, the first one is returned.

        Args:
            name: Section name.

        Returns:
            Section content as a list of lines.

        """
        for header, section in self.sections:
            if header.lower() == name.lower():
                return section
        return None

    def replace_section(self, name: str, content: List[str]) -> bool:
        """Replaces content of a section.

        In case there are multiple sections with the same name, the first one is replaced.

        Args:
            name: Section name.
            content: Section content as a list of lines.

        Returns:
            False if section was not found else True.

        """
        for i, (header, _) in enumerate(self.sections):
            if header.lower() == name.lower():
                self.sections[i] = (header, content)
                return True
        return False

    @classmethod
    def _split_sections(cls, content: str) -> List[Tuple[str, List[str]]]:
        """Splits content of a SPEC file into sections.

        Args:
            content: Content of the SPEC file

        Returns:
            The split sections represented as a list of tuples (the first element
            is the section name, the second is the content of the section).

        """
        lines = content.splitlines()
        section_headers_re = [re.compile(r'^{0}\b.*'.format(re.escape(x)), re.IGNORECASE) for x in cls.SECTION_HEADERS]

        section_beginnings: List[Optional[int]] = []
        for i, line in enumerate(lines):
            if line.startswith('%'):
                for header in section_headers_re:
                    if header.match(line):
                        section_beginnings.append(i)
        section_beginnings.append(None)

        sections = [('%package', lines[:section_beginnings[0]])]

        for i in range(len(section_beginnings) - 1):
            # Only the last element of section_beginnings is None, section_beginnings[i] can never be None
            # (the last element isn't iterated over), so section_beginnings[i] + 1 is always valid. Ignore the type.
            start = section_beginnings[i] + 1  # type: ignore
            end = section_beginnings[i + 1]
            sections.append((lines[start - 1], lines[start:end]))
        return sections



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re
import sys

from rebasehelper.cli import CLI
from rebasehelper.archive import Archive


class Completion:

    @staticmethod
    def extensions():
        archives = Archive.get_supported_archives()
        return [a.lstrip('.') for a in archives]

    @staticmethod
    def options():
        def get_delimiter(parser, action):
            if action.nargs == 0:
                return None
            fmt = parser._get_formatter()  # pylint: disable=protected-access
            usage = fmt._format_actions_usage([action], [])  # pylint: disable=protected-access
            option_string = action.option_strings[0]
            idx = usage.find(option_string)
            if idx == -1:
                return None
            return usage[idx + len(option_string)]
        parser = CLI.build_parser()
        result = []
        actions = parser._get_optional_actions() + parser._get_positional_actions()  # pylint: disable=protected-access
        for action in actions:
            if not action.option_strings:
                continue
            delimiter = get_delimiter(parser, action) or ''
            result.append(dict(
                options=[o + delimiter.strip() for o in action.option_strings],
                choices=action.choices or []))
        return result

    @classmethod
    def dump(cls):
        options = cls.options()
        return {
            # pattern list of extensions
            'RH_EXTENSIONS': '@({})'.format('|'.join(cls.extensions())),
            # array of options
            'RH_OPTIONS': '({})'.format(' '.join('"{}"'.format(' '.join(o['options'])) for o in options)),
            # array of choices of respective options
            'RH_CHOICES': '({})'.format(' '.join('"{}"'.format(' '.join(o['choices'])) for o in options)),
        }


def replace_placeholders(s, **kwargs):
    placeholder_re = re.compile(r'@(\w+)@')
    matches = list(placeholder_re.finditer(s))
    result = s
    for match in reversed(matches):
        replacement = kwargs.get(match.group(1), '')
        result = result[:match.start(0)] + replacement + result[match.end(0):]
    return result


def main():
    if len(sys.argv) != 3:
        return 1
    with open(sys.argv[1]) as f:
        s = f.read()
    s = replace_placeholders(s, **Completion.dump())
    with open(sys.argv[2], 'w') as f:
        f.write(s)
    return 0


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import shlex

from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection


class BuildToolBase(Plugin):
    """Build tool base class.

    Attributes:
        DEFAULT(bool): If True, the build tool is default tool.
        ACCEPTS_OPTIONS(bool): If True, the build tool accepts additional
            options passed via --builder-options.
        CREATES_TASKS(bool): If True, the build tool creates remote tasks.

    """

    DEFAULT: bool = False
    ACCEPTS_OPTIONS: bool = False
    CREATES_TASKS: bool = False

    @classmethod
    def prepare(cls, spec, conf):
        """
        Prepare for building.

        :param spec: spec file object
        """
        # do nothing by default

    @classmethod
    def build(cls, spec, results_dir, srpm, **kwargs):
        """
        Build binaries from the sources.

        Keyword arguments:
        spec -- path to a SPEC file
        sources -- list with absolute paths to SOURCES
        patches -- list with absolute paths to PATCHES
        results_dir -- path to DIR where results should be stored

        Returns:
        dict with:
        'srpm' -> absolute path to SRPM
        'rpm' -> list of absolute paths to RPMs
        'logs' -> list of absolute paths to logs
        """
        raise NotImplementedError()

    @classmethod
    def wait_for_task(cls, build_dict, task_id, results_dir):  # pylint: disable=unused-argument
        """
        Waits until specified task is finished

        :param build_dict: build data
        :param results_dir: path to DIR where results should be stored
        :return: tuple with:
            list of absolute paths to RPMs
            list of absolute paths to logs
        """
        # do nothing by default
        return build_dict.get('rpm'), build_dict.get('logs')

    @classmethod
    def get_task_info(cls, build_dict):
        """
        Gets information about detached remote task

        :param build_dict: build data
        :return: task info
        """
        raise NotImplementedError()

    @classmethod
    def get_detached_task(cls, task_id, results_dir):
        """
        Gets packages and logs for specified task

        :param task_id: detached task id
        :param results_dir: path to DIR where results should be stored
        :return: tuple with:
            list of absolute paths to RPMs
            list of absolute paths to logs
        """
        raise NotImplementedError()

    @staticmethod
    def get_builder_options(**kwargs):
        builder_options = kwargs.get('builder_options')
        if builder_options:
            return shlex.split(builder_options)
        return None


class BuildToolCollection(PluginCollection):
    """Collection of RPM build tools."""



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import time
import urllib.parse
from typing import cast

import pyquery  # type: ignore
from copr.v3 import Client, CoprConfigException, CoprNoConfigException, CoprRequestException  # type: ignore

from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.helpers.download_helper import DownloadHelper
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class CoprHelper:

    DELETE_PROJECT_AFTER = 60

    @classmethod
    def get_client(cls):
        try:
            client = Client.create_from_config_file()
        except (CoprNoConfigException, CoprConfigException) as e:
            raise RebaseHelperError('Missing or invalid copr configuration file') from e
        else:
            return client

    @classmethod
    def create_project(cls, client, project, chroots, description, instructions, permanent=False, hide=True):
        try:
            client.project_proxy.add(ownername=client.config.get('username'),
                                     projectname=project,
                                     chroots=chroots,
                                     delete_after_days=None if permanent else cls.DELETE_PROJECT_AFTER,
                                     unlisted_on_hp=hide,
                                     description=description,
                                     instructions=instructions)
        except CoprRequestException as e:
            error = e.result.error
            try:
                [[error]] = error.values()
            except AttributeError:
                pass

            if error.startswith('You already have project named'):
                # reuse existing project
                pass
            else:
                raise RebaseHelperError('Failed to create copr project. Reason: {}'.format(error)) from e

    @classmethod
    def build(cls, client, project, srpm):
        try:
            result = client.build_proxy.create_from_file(client.config.get('username'), project, srpm)
        except CoprRequestException as e:
            raise RebaseHelperError('Failed to start copr build: {}'.format(str(e))) from e
        else:
            return result.id

    @classmethod
    def get_build_url(cls, client, build_id):
        try:
            result = client.build_proxy.get(build_id)
        except CoprRequestException as e:
            raise RebaseHelperError(
                'Failed to get copr build details for id {}: {}'.format(build_id, str(e))) from e
        else:
            return '{}/coprs/{}/{}/build/{}/'.format(client.config.get('copr_url'),
                                                     result.ownername,
                                                     result.projectname,
                                                     build_id)

    @classmethod
    def get_build_status(cls, client, build_id):
        try:
            result = client.build_proxy.get(build_id)
        except CoprRequestException as e:
            raise RebaseHelperError(
                'Failed to get copr build details for id {}: {}'.format(build_id, str(e))) from e
        else:
            return result.state

    @classmethod
    def watch_build(cls, client, build_id):
        try:
            logged = False
            while True:
                status = cls.get_build_status(client, build_id)
                if not status:
                    return False
                elif status in ['succeeded', 'skipped']:
                    return True
                elif status in ['failed', 'canceled', 'unknown']:
                    return False
                else:
                    if not logged:
                        logger.info('Waiting for copr build to finish')
                        logged = True
                    time.sleep(10)
        except KeyboardInterrupt:
            return False

    @classmethod
    def download_build(cls, client, build_id, destination):
        logger.info('Downloading packages and logs for build %d', build_id)
        rpms = []
        logs = []
        for chroot in client.build_chroot_proxy.get_list(build_id):
            url = chroot.result_url
            url = url if url.endswith('/') else url + '/'
            d = pyquery.PyQuery(url)
            d.make_links_absolute()
            for a in d('a[href$=\'.rpm\'], a[href$=\'.log.gz\']'):
                fn = os.path.basename(urllib.parse.urlsplit(a.attrib['href']).path)
                dest = os.path.join(destination, chroot.name)
                os.makedirs(dest, exist_ok=True)
                dest = os.path.join(dest, fn)
                if fn.endswith('.src.rpm'):
                    # skip source RPM
                    continue
                DownloadHelper.download_file(a.attrib['href'], dest)
                if fn.endswith('.rpm'):
                    rpms.append(dest)
                elif fn.endswith('.log.gz'):
                    local_path = dest.replace('.log.gz', '.log')
                    os.rename(dest, local_path)
                    logs.append(local_path)
        return rpms, logs



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
from typing import Dict, Optional, cast

from rebasehelper.exceptions import RebaseHelperError, CheckerNotFoundError
from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.rpm_helper import RpmHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))

AbiChange = Dict[str, Dict[str, Dict[str, str]]]


class AbiPkgDiff(BaseChecker):
    """abipkgdiff compare tool."""

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.RPM
    results_dir: Optional[str] = ''

    CMD: str = 'abipkgdiff'
    ABIDIFF_ERROR: int = 1 << 0
    ABIDIFF_USAGE_ERROR: int = 1 << 1
    ABIDIFF_ABI_CHANGE: int = 1 << 2
    ABIDIFF_ABI_INCOMPATIBLE_CHANGE: int = 1 << 3

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 3
        except (IOError, OSError):
            return False

    # Example
    # abipkgdiff --d1 dbus-glib-debuginfo-0.80-3.fc12.x86_64.rpm \
    # --d2 dbus-glib-debuginfo-0.104-3.fc23.x86_64.rpm \
    # dbus-glib-0.80-3.fc12.x86_64.rpm dbus-glib-0.104-3.fc23.x86_64.rpm
    @classmethod
    def _get_packages_for_abipkgdiff(cls, input_structure=None):
        debug_package = None
        rest_packages = None
        packages = input_structure.get('rpm', [])
        if packages:
            debug_package = [x for x in packages if 'debuginfo' in os.path.basename(x)]
            rest_packages = [x for x in packages if 'debuginfo' not in os.path.basename(x)]

        return debug_package, rest_packages

    @classmethod
    def _find_debuginfo(cls, debug, pkg):
        name = RpmHelper.split_nevra(os.path.basename(pkg))['name']
        debuginfo = '{}-debuginfo'.format(name)
        find = [x for x in debug if RpmHelper.split_nevra(os.path.basename(x))['name'] == debuginfo]
        if find:
            return find[0]
        srpm = RpmHelper.get_header_from_rpm(pkg).sourcerpm
        debuginfo = '{}-debuginfo'.format(RpmHelper.split_nevra(srpm)['name'])
        find = [x for x in debug if RpmHelper.split_nevra(os.path.basename(x))['name'] == debuginfo]
        if find:
            return find[0]
        return None

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """Compares old and new RPMs using abipkgdiff"""
        # Check if ABI changes occured
        cls.results_dir = os.path.join(results_dir, cls.name)
        cls.prepare_results_dir()
        debug_old, rest_pkgs_old = cls._get_packages_for_abipkgdiff(results_store.get_build('old'))
        debug_new, rest_pkgs_new = cls._get_packages_for_abipkgdiff(results_store.get_build('new'))
        ret_codes = {}
        for pkg in rest_pkgs_old:
            command = [cls.CMD]
            command_fallback = [cls.CMD]
            debug = cls._find_debuginfo(debug_old, pkg)
            if debug:
                command.append('--d1')
                command.append(debug)
            old_name = RpmHelper.split_nevra(os.path.basename(pkg))['name']
            find = [x for x in rest_pkgs_new if RpmHelper.split_nevra(os.path.basename(x))['name'] == old_name]
            if not find:
                logger.warning('New version of package %s was not found!', old_name)
                continue
            new_pkg = find[0]
            debug = cls._find_debuginfo(debug_new, new_pkg)
            if debug:
                command.append('--d2')
                command.append(debug)
            command.extend([pkg, new_pkg])
            command_fallback.extend([pkg, new_pkg])
            logger.verbose('Package name for ABI comparison %s', old_name)
            output = os.path.join(cls.results_dir, old_name + '.txt')
            for cmd in [command, command_fallback]:
                try:
                    ret_code = ProcessHelper.run_subprocess(cmd, output_file=output)
                except OSError as e:
                    raise CheckerNotFoundError("Checker '{}' was not found or installed.".format(cls.name)) from e
                if int(ret_code) & cls.ABIDIFF_ERROR and int(ret_code) & cls.ABIDIFF_USAGE_ERROR:
                    raise RebaseHelperError(
                        'Execution of {} failed.\nCommand line is: {}'.format(cls.CMD, ' '.join(cmd)))
                if int(ret_code) & cls.ABIDIFF_ERROR:
                    # abipkgdiff might not be able to read the debuginfo, try again without it
                    continue
                break
            ret_codes[old_name] = int(ret_code)
        return dict(packages=cls.parse_abi_logs(ret_codes),
                    abi_changes=any(x & cls.ABIDIFF_ABI_CHANGE for x in ret_codes.values()),
                    abi_incompatible_changes=any(x & cls.ABIDIFF_ABI_INCOMPATIBLE_CHANGE for x in ret_codes.values()),
                    path=cls.get_checker_output_dir_short(),
                    ret_codes=ret_codes)

    @classmethod
    def parse_abi_logs(cls, ret_codes):
        """Parses summary information from abipkgdiff logs.

        Args:
            ret_codes(dict): Dictionary mapping package names to abipkgdiff return codes.

        Returns:
            dict: Dictionary mapping package names to a dict of summary information for each shared object.

            For example:

            {
                'libtiff':
                {
                   'libtiff.so.5.2.5':{
                        'Functions changes summary': {
                            'Added': {
                                'count': '8',
                                'what': 'Added',
                                'filtered_out': None,
                            },
                        },
                        'Variables changes summary': {},
                    },
                },
            }
        """
        def parse_changes(lines):
            title_re = re.compile(r"^\s*=+\s*changes\s+of\s+'(?P<filename>.+)'.*$")
            summary_re = re.compile(r'''^
            \s+(?P<kind>[\w\s]+changes\s+summary):\s+
            (?P<changes>.+)
            $
            ''', re.VERBOSE)
            changes_re = re.compile(r'''
            (?P<count>\d+)\s+
            (?P<what>Added|Changed|Removed)(\s+functions|variables)?
            (\s+\((?P<filtered_out>\d+)\s+filtered\s+out\))?
            ''', re.VERBOSE)
            result_dict: Dict[str, AbiChange] = {}
            filename = 'Undetected filename'
            for line in lines:
                mt = title_re.match(line)
                if mt:
                    filename = mt.group('filename')
                    if filename not in result_dict:
                        result_dict[filename] = {}
                ms = summary_re.match(line)
                if ms:
                    result: AbiChange = {}
                    ds = ms.groupdict()
                    result[ds['kind']] = {}
                    for mc in changes_re.finditer(ds['changes']):
                        dc = mc.groupdict()
                        dc['count'] = int(dc['count'])
                        if int(dc['count']) or dc['filtered_out']:
                            result[ds['kind']][dc['what']] = dc

                    result_dict[filename].update(result)
            return result_dict

        pkgs = {}
        for pkg, ret_code in ret_codes.items():
            if ret_code & cls.ABIDIFF_ABI_CHANGE:
                with open(os.path.join(cls.results_dir, pkg + '.txt'), 'r') as f:
                    pkgs[pkg] = parse_changes(f.readlines())
        return pkgs

    @classmethod
    def format(cls, data):
        output_lines = [cls.get_underlined_title('abipkgdiff')]
        if not cls.get_important_changes(data):
            output_lines.append('No ABI changes occured.')
            return output_lines
        for pkg_name, pkg_changes in sorted(data['packages'].items()):
            if not pkg_changes:
                continue
            output_lines.append("ABI changes in {}:".format(pkg_name))
            for filename, file_changes in pkg_changes.items():
                output_lines.append(" - {}:".format(filename))
                for sum_title, changes_list in sorted(file_changes.items()):
                    if not changes_list:
                        continue
                    output_lines.append("   - {}:".format(sum_title))

                    for change_name, change_info in sorted(changes_list.items()):
                        if change_info['filtered_out']:
                            output_lines.append("     - {} {} (filtered out {})".format(change_name,
                                                                                        change_info['count'],
                                                                                        change_info['filtered_out']))
                        else:
                            output_lines.append("     - {} {}".format(change_name, change_info['count']))
        output_lines.append("Details in {}:".format(data['path']))
        for pkg_name, pkg_changes in sorted(data['packages'].items()):
            output_lines.append(" - {}.txt".format(pkg_name))

        return output_lines

    @classmethod
    def get_important_changes(cls, checker_output):
        if checker_output['abi_incompatible_changes']:
            return ['Incompatible ABI changes occurred. Check abipkgdiff output.']
        if checker_output['abi_changes']:
            return ['ABI changes occurred. Check abipkgdiff output.']
        return []
